# Copyright (c) 2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Text generation.
"""

import argparse
import sys

import pyeddl.eddl as eddl
from pyeddl.tensor import Tensor


MEM_CHOICES = ("low_mem", "mid_mem", "full_mem")


def ResBlock(layer, filters, nconv, half):
    in_ = layer

    if half:
        layer = eddl.ReLu(eddl.BatchNormalization(
            eddl.Conv(layer, filters, [3, 3], [2, 2]), True
        ))
    else:
        layer = eddl.ReLu(eddl.BatchNormalization(
            eddl.Conv(layer, filters, [3, 3], [1, 1]), True
        ))

    for i in range(nconv - 1):
        layer = eddl.ReLu(eddl.BatchNormalization(
            eddl.Conv(layer, filters, [3, 3], [1, 1]), True
        ))

    if half:
        return eddl.Add(eddl.BatchNormalization(
            eddl.Conv(in_, filters, [1, 1], [2, 2]), True
        ), layer)
    else:
        return eddl.Add(layer, in_)


def main(args):
    eddl.download_flickr()

    olength = 20
    outvs = 2000
    embdim = 32

    # Define network
    in_ = eddl.Input([3, 256, 256])  # Image
    layer = in_
    layer = eddl.ReLu(eddl.Conv(layer, 64, [3, 3], [2, 2]))

    layer = ResBlock(layer, 64, 2, 1)
    layer = ResBlock(layer, 64, 2, 0)
    layer = ResBlock(layer, 128, 2, 1)
    layer = ResBlock(layer, 128, 2, 0)
    layer = ResBlock(layer, 256, 2, 1)
    layer = ResBlock(layer, 256, 2, 0)
    layer = ResBlock(layer, 512, 2, 1)
    layer = ResBlock(layer, 512, 2, 0)
    layer = eddl.GlobalAveragePool(layer)
    layer = eddl.Reshape(layer, [-1])

    # Decoder
    ldec = eddl.Input([outvs])
    ldec = eddl.ReduceArgMax(ldec, [0])
    ldec = eddl.RandomUniform(
        eddl.Embedding(ldec, outvs, 1, embdim), -0.05, 0.05
    )
    layer = eddl.Decoder(eddl.LSTM(ldec, 512, True), layer, "concat")
    out = eddl.Softmax(eddl.Dense(layer, outvs))
    net = eddl.Model([in_], [out])

    # Build model
    eddl.build(
        net,
        eddl.adam(0.001),
        ["softmax_cross_entropy"],
        ["accuracy"],
        eddl.CS_GPU(mem=args.mem) if args.gpu else eddl.CS_CPU(mem=args.mem)
    )
    eddl.summary(net)

    # Load dataset
    x_train = Tensor.load("flickr_trX.bin")
    y_train = Tensor.load("flickr_trY.bin")
    if args.small:
        x_train = x_train.select([":200"])
        y_train = y_train.select([":200"])
    xtrain = Tensor.permute(x_train, [0, 3, 1, 2])
    y_train = Tensor.onehot(y_train, outvs)
    # batch x timesteps x input_dim
    y_train.reshape_([y_train.shape[0], olength, outvs])

    # Train model
    for i in range(args.epochs):
        eddl.fit(net, [xtrain], [y_train], args.batch_size, 1)

    print("All done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--epochs", type=int, metavar="INT", default=10)
    parser.add_argument("--batch-size", type=int, metavar="INT", default=24)
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--small", action="store_true")
    parser.add_argument("--mem", metavar="|".join(MEM_CHOICES),
                        choices=MEM_CHOICES, default="low_mem")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from urllib.request import urlretrieve

import numpy as np
from pyeddl.tensor import Tensor

# Convert array to tensor and save to "bin" format
a = np.arange(6).reshape([2, 3]).astype(np.float32)
print(a)
t = Tensor.fromarray(a)
t.save("./a.bin", "bin")
t1 = Tensor.load("a.bin", "bin")
a1 = t1.getdata()
print(a1)

print()

# Read numpy data and convert to tensors
FNAME = "mnist.npz"
LOC = "https://storage.googleapis.com/tensorflow/tf-keras-datasets"
if not os.path.exists(FNAME):
    fname, _ = urlretrieve("%s/%s" % (LOC, FNAME), FNAME)
    print("Downloaded", fname)
print("loading", FNAME)
with np.load(FNAME) as f:
    x_train, y_train = f['x_train'], f['y_train']
    x_test, y_test = f['x_test'], f['y_test']
t_x_train = Tensor.fromarray(x_train.astype(np.float32))
t_y_train = Tensor.fromarray(y_train.astype(np.float32))
t_x_test = Tensor.fromarray(x_test.astype(np.float32))
t_y_test = Tensor.fromarray(y_test.astype(np.float32))
print(x_train.shape, x_train.dtype)
t_x_train.info()



# Copyright (c) 2019-2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest
import pyeddl.eddl as eddl
from pyeddl.tensor import Tensor
from pyeddl._core import Loss, Metric


class MSELoss(Loss):

    def __init__(self):
        Loss.__init__(self, "py_mean_squared_error")

    def delta(self, t, y, d):
        Tensor.copy(Tensor.sub(y, t), d)
        d.div_(t.shape[0])

    def value(self, t, y):
        size = t.size / t.shape[0]
        return (Tensor.sqr(Tensor.sub(t, y))).sum() / size


class MSEMetric(Metric):

    def __init__(self):
        Metric.__init__(self, "py_mean_squared_error")

    def value(self, t, y):
        size = t.size / t.shape[0]
        return (Tensor.sqr(Tensor.sub(t, y))).sum() / size


class CategoricalAccuracy(Metric):

    def __init__(self):
        Metric.__init__(self, "py_categorical_accuracy")

    def value(self, t, y):
        a = t.getdata()
        b = y.getdata()
        return (np.argmax(a, axis=-1) == np.argmax(b, axis=-1)).sum()


def test_py_metric():
    shape = [8, 10]
    a = np.random.random(shape).astype(np.float32)
    b = np.random.random(shape).astype(np.float32)
    t, y = Tensor.fromarray(a), Tensor.fromarray(b)
    v = MSEMetric().value(t, y)
    exp_v = eddl.getMetric("mse").value(t, y)
    assert v == pytest.approx(exp_v)
    v = CategoricalAccuracy().value(t, y)
    exp_v = eddl.getMetric("categorical_accuracy").value(t, y)
    assert v == pytest.approx(exp_v)


def test_py_loss():
    shape = [8, 10]
    a = np.random.random(shape).astype(np.float32)
    b = np.random.random(shape).astype(np.float32)
    t, y = Tensor.fromarray(a), Tensor.fromarray(b)
    z = Tensor(shape)
    exp_z = Tensor(shape)
    py_mse_loss = MSELoss()
    mse_loss = eddl.getLoss("mse")
    mse_loss.delta(t, y, exp_z)
    py_mse_loss.delta(t, y, z)
    c = np.array(z, copy=False)
    exp_c = np.array(exp_z, copy=False)
    assert np.array_equal(c, exp_c)
    v = py_mse_loss.value(t, y)
    exp_v = mse_loss.value(t, y)
    assert v == pytest.approx(exp_v)



# Copyright (c) 2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
CIFAR example from pretrained resnet18.
"""

import argparse
import os
import sys
from urllib.request import urlretrieve

import pyeddl.eddl as eddl
from pyeddl.tensor import Tensor


MEM_CHOICES = ("low_mem", "mid_mem", "full_mem")
MODEL_URL = "https://www.dropbox.com/s/tn0d87dr035yhol/resnet18-v1-7.onnx"
MODEL_PATH = "resnet18-v1-7.onnx"


def main(args):
    eddl.download_cifar10()
    if not os.path.isfile(MODEL_PATH):
        print("downloading", os.path.basename(MODEL_PATH))
        urlretrieve(MODEL_URL + "?dl=1", MODEL_PATH)

    net_onnx = eddl.import_net_from_onnx_file(MODEL_PATH)
    # remove last layer
    eddl.removeLayer(net_onnx, "resnetv15_dense0_fwd")

    # create a new graph to adapt the output for CIFAR
    in_ = eddl.Input([512])
    layer = eddl.Dense(in_, 10)
    layer = eddl.Softmax(layer)

    net_adap = eddl.Model([in_], [layer])

    # cat both models
    net = eddl.Model([net_onnx, net_adap])

    eddl.build(
        net,
        eddl.sgd(0.001),
        ["softmax_cross_entropy"],
        ["accuracy"],
        eddl.CS_GPU(mem=args.mem) if args.gpu else eddl.CS_CPU(mem=args.mem)
    )
    eddl.summary(net)

    # from "flatten_170" to the beginning
    eddl.setTrainable(net, "flatten_170", False)

    # Load dataset
    x_train = Tensor.load("cifar_trX.bin")
    xs = x_train.select([":1000", ":"])
    xtrain = xs.scale([224, 224])
    xtrain.div(255.0)
    del x_train
    del xs

    y_train = Tensor.load("cifar_trY.bin")
    ytrain = y_train.select([":1000", ":"])

    for i in range(args.epochs):
        eddl.fit(net, [xtrain], [ytrain], args.batch_size, 1)

    print("All done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--epochs", type=int, metavar="INT", default=10)
    parser.add_argument("--batch-size", type=int, metavar="INT", default=100)
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--small", action="store_true")  # no-op
    parser.add_argument("--mem", metavar="|".join(MEM_CHOICES),
                        choices=MEM_CHOICES, default="low_mem")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import _core
_eddl = _core.eddl


# == MODEL METHODS ==

# = Creation =

def Model(in_, out=None):
    r"""\
    Create a model (Net).

    With one argument, create from a list of models. With two arguments,
    create from a list of input layers and a list of output layers.

    :param in\_: if this is the only parameter, it must be a list of models.
      If ``out`` is specified, this must be a list of input layers.
    :param out: list of output layers
    :return: model instance
    """
    if out is None:
        return _eddl.Model(in_)
    return _eddl.Model(in_, out)


def setName(m, name):
    return _eddl.setName(m, name)


def getLayer(net, in_):
    # note: in_ can be a vector of layers or a string
    return _eddl.getLayer(net, in_)


def removeLayer(net, l):
    _eddl.removeLayer(net, l)


def get_parameters(net, deepcopy=False, tocpu=False):
    return _eddl.get_parameters(net, deepcopy, tocpu)


def set_parameters(net, params):
    _eddl.set_parameters(net, params)


def build(net, o=None, lo=None, me=None, cs=None, init_weights=True):
    """\
    Tell the model which optimizer, losses, metrics and computing service to
    use.

    :param net: model
    :param o: optimizer
    :param lo: list of losses
    :param me: list of metrics
    :param cs: computing service
    :return: None
    """
    # core module has multiple overloads for this:
    #  1. build(net, o=None, cs=None, init_weigths=True)
    #  2. build(net, o, lo, me, cs=None, init_weights=True)
    if (lo is None and me is not None) or (lo is not None and me is None):
        raise ValueError("lo and me must be both None or both not None")
    if lo is None:
        return _eddl.build(net, o, cs, init_weights)  # overload 1
    if o is None:
        raise ValueError("if lo and me are provided, o cannot be None")
    return _eddl.build(net, o, lo, me, cs, init_weights)  # overload 2


# = Computing services =

def toGPU(net, g=None, lsb=None, mem=None):
    """\
    Assign model operations to the GPU.

    :param net: model
    :param g: list of gpu ids to allocate
    :param lsb: number of batches to sync model weights
    :return: None
    """
    # core module has multiple overloads for this:
    #  1. toGPU(net, g, lsb)
    #  2. toGPU(net, g, mem)
    #  3. toGPU(net, g, lsb, mem)
    #  4. toGPU(net, g)
    #  5. toGPU(net, mem)
    #  6. toGPU(net)
    if g is None:
        if mem is None:
            return _eddl.toGPU(net)
        return _eddl.toGPU(net, mem)
    if lsb is None:
        return _eddl.toGPU(net, g, mem)
    if mem is None:
        return _eddl.toGPU(net, g, lsb)
    return _eddl.toGPU(net, g, lsb, mem)


def toCPU(net, t=None):
    """\
    Assign model operations to the CPU.

    :param net: model
    :param t: number of CPU threads
    :return: None
    """
    if t is None:
        return _eddl.toCPU(net)
    return _eddl.toCPU(net, t)


def CS_CPU(th=-1, mem="full_mem"):
    """\
    Create a computing service that executes the code in the CPU.

    :param th: number of threads to use (-1 = all available threads)
    :param mem: memory consumption of the model: "full_mem", "mid_mem" or
      "low_mem"
    :return: computing service
    """
    return _eddl.CS_CPU(th, mem)


def CS_GPU(g=[1], lsb=1, mem="full_mem"):
    """\
    Create a computing service that executes the code in the GPU.

    :param g: list of integers to set which GPUs will be used (1=on, 0=off)
    :param lsb: (multi-gpu setting) number of batches to run before
      synchronizing the weights of the different GPUs
    :param mem: memory consumption of the model: "full_mem", "mid_mem" or
      "low_mem"
    :return: computing service
    """
    return _eddl.CS_GPU(g, lsb, mem)


def CS_FPGA(f, lsb=1):
    """\
    Create a computing service that executes the code in the FPGA.

    :param f: list of integers to set which FPGAs will be used (1=on, 0=off)
    :param lsb: (multi-fpga setting) number of batches to run before
      synchronizing the weights of the different FPGAs
    :return: computing service
    """
    return _eddl.CS_FPGA(f, lsb)


def CS_COMPSS(filename):
    """\
    Create a computing service that executes the code in the COMPSs framework.

    :param filename: file with the setup specification
    :return: computing service
    """
    return _eddl.CS_COMPSS(filename)


# = Info and logs =

def setlogfile(net, fname):
    """\
    Save the training outputs of a model to a file.

    :param net: model
    :param fname: name of the log file
    :return: None
    """
    return _eddl.setlogfile(net, fname)


def summary(m):
    """\
    Print a summary representation of the model.

    :param m: model
    :return: None
    """
    return _eddl.summary(m)


def plot(m, fname, string="LR"):
    """\
    Plot a representation of the model.

    :param m: model to plot
    :param fname: name of the file where the plot should be saved
    :return: None
    """
    return _eddl.plot(m, fname, string)


# = Serialization =

def load(m, fname, format="bin"):
    """\
    Load weights to reinstantiate the model.

    :param m: model
    :param fname: name of the file containing the model weights
    :return: None
    """
    return _eddl.load(m, fname, format)


def save(m, fname, format="bin"):
    """\
    Save model weights to a file.

    :param m: model
    :param fname: name of the file where model weights should be saved
    :return: None
    """
    return _eddl.save(m, fname, format)


# = Optimizer =

def setlr(net, p):
    """\
    Change the learning rate and hyperparameters of the model optimizer.

    :param net: model
    :param p: list with the learning rate and hyperparameters of the model
    :return: None
    """
    return _eddl.setlr(net, p)


def adadelta(lr, rho, epsilon, weight_decay):
    """\
    Adadelta optimizer.

    Adadelta is a more robust extension of Adagrad that adapts learning rates
    based on a moving window of gradient updates, instead of accumulating all
    past gradients. This way, Adadelta continues learning even when many
    updates have been done. See: https://arxiv.org/abs/1212.5701.

    :param lr: learning rate
    :param rho: smoothing constant
    :param epsilon: term added to the denominator to improve numerical
      stability
    :param weight_decay: weight decay (L2 penalty)
    :return: Adadelta optimizer
    """
    return _eddl.adadelta(lr, rho, epsilon, weight_decay)


def adam(lr=0.01, beta_1=0.9, beta_2=0.999, epsilon=0.000001, weight_decay=0,
         amsgrad=False):
    """\
    Adam optimizer.

    Default parameters follow those provided in the original paper
    (see: https://arxiv.org/abs/1412.6980v8).

    :param lr: learning rate
    :param beta_1: coefficient for computing running averages of gradient
      and its square
    :param beta_2: coefficient for computing running averages of gradient
      and its square
    :param epsilon: term added to the denominator to improve numerical
      stability
    :param weight_decay: weight decay (L2 penalty)
    :param amsgrad: whether to apply the AMSGrad variant of this algorithm
      from the paper "On the Convergence of Adam and Beyond"
    :return: Adam optimizer
    """
    return _eddl.adam(lr, beta_1, beta_2, epsilon, weight_decay, amsgrad)


def adagrad(lr, epsilon, weight_decay):
    """\
    Adagrad optimizer.

    Adagrad is an optimizer with parameter-specific learning rates, which are
    adapted relative to how frequently a parameter gets updated during
    training. The more updates a parameter receives, the smaller the learning
    rate. See: http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf

    :param lr: learning rate
    :param epsilon: term added to the denominator to improve numerical
      stability
    :param weight_decay: weight decay (L2 penalty)
    :return: Adagrad optimizer
    """
    return _eddl.adagrad(lr, epsilon, weight_decay)


def adamax(lr, beta_1, beta_2, epsilon, weight_decay):
    """\
    Adamax optimizer.

    A variant of Adam based on the infinity norm.

    :param lr: learning rate
    :param beta_1: coefficient for computing running averages of gradient
      and its square
    :param beta_2: coefficient for computing running averages of gradient
      and its square
    :param epsilon: term added to the denominator to improve numerical
      stability
    :param weight_decay: weight decay (L2 penalty)
    :return: Adamax optimizer
    """
    return _eddl.adamax(lr, beta_1, beta_2, epsilon, weight_decay)


def nadam(lr, beta_1, beta_2, epsilon, schedule_decay):
    """\
    Nadam optimizer.

    :param lr: learning rate
    :param beta_1: coefficients for computing running averages of gradient
      and its square
    :param beta_2: coefficients for computing running averages of gradient
      and its square
    :param epsilon: term added to the denominator to improve numerical
      stability
    :param schedule_decay: weight decay (L2 penalty)
    :return: Nadam optimizer
    """
    return _eddl.nadam(lr, beta_1, beta_2, epsilon, schedule_decay)


def rmsprop(lr=0.01, rho=0.9, epsilon=0.00001, weight_decay=0.0):
    """\
    RMSProp optimizer.

    It is recommended to leave the parameters of this optimizer at their
    default values (except for the learning rate, which can be freely tuned).
    See:
    http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf

    :param lr: learning rate
    :param rho: smoothing constant
    :param epsilon: term added to the denominator to improve numerical
      stability
    :param weight_decay: weight decay (L2 penalty)
    :return: RMSProp optimizer
    """
    return _eddl.rmsprop(lr, rho, epsilon, weight_decay)


def sgd(lr=0.01, momentum=0.0, weight_decay=0.0, nesterov=False):
    """\
    Stochastic gradient descent optimizer.

    Includes support for momentum, learning rate decay, and Nesterov momentum.

    :param lr: learning rate
    :param momentum: momentum factor
    :param weight_decay: value to apply to the activation function
    :param nesterov: whether to apply Nesterov momentum
    :return: SGD optimizer
    """
    return _eddl.sgd(lr, momentum, weight_decay, nesterov)


# Training and evaluation - coarse methods

def fit(m, in_, out, batch, epochs):
    """\
    Train the model for a fixed number of epochs (iterations on a dataset).

    :param m: model to train
    :param in_: input data (features)
    :param out: output data (labels)
    :param batch: number of samples per gradient update
    :param epochs: number of training epochs. An epoch is an iteration over
      the entire data provided
    :return: None
    """
    return _eddl.fit(m, in_, out, batch, epochs)


def evaluate(m, in_, out, bs=100):
    """\
    Compute the loss and metric values for the model in test mode.

    :param m: model to train
    :param in_: input data (features)
    :param out: output data (labels)
    :param bs: batch size
    :return: None
    """
    return _eddl.evaluate(m, in_, out, bs)


def predict(m, in_):
    r"""\
    Perform a prediction with input data

    :param m: model
    :param in\_: input data (features)
    :return: output tensors
    """
    return _eddl.predict(m, in_)


# Training and evaluation - finer methods

def random_indices(batch_size, num_samples):
    """\
    Generate a random sequence of indices for a batch.

    :param batch_size: length of the random sequence to generate
    :param num_samples: number of samples available, i.e., maximum value to
      include in the random sequence + 1
    :return: list of integers
    """
    return _eddl.random_indices(batch_size, num_samples)


def train_batch(net, in_, out, indices=None):
    """\
    Train the model using the samples of the input list that are on the
    selected indices list.

    :param net: model to train
    :param in_: list of samples
    :param out: list of labels or expected output
    :param indices: list of indices of the samples to train
    :return: None
    """
    if indices is None:
        return _eddl.train_batch(net, in_, out)
    return _eddl.train_batch(net, in_, out, indices)


def eval_batch(net, in_, out, indices=None):
    """\
    Evaluate the model using the samples of the input list that are on the
    selected indices list.

    :param net: model to evaluate
    :param in_: list of samples
    :param out: list of labels or expected output
    :param indices: list of indices of the samples to train
    :return: None
    """
    if indices is None:
        return _eddl.eval_batch(net, in_, out)
    return _eddl.eval_batch(net, in_, out, indices)


def next_batch(in_, out):
    """\
    Load the next batch of random samples from the input list to the
    output list.

    :param in_: list from where the samples of the next batch should be
      chosen from
    :param out: list where the samples of the next batch should be stored
    :return: None
    """
    return _eddl.next_batch(in_, out)


# Training and evaluation - finest methods

def set_mode(net, mode):
    """\
    Set model mode.

    :param net: model
    :param mode: 1 for training, 0 for test
    :return: None
    """
    return _eddl.set_mode(net, mode)


def reset_loss(m):
    """\
    Reset model loss.

    :param m: model
    :return: None
    """
    return _eddl.reset_loss(m)


def forward(m, in_=None, b=None):
    """\
    Compute the gradient of the model through the forward graph

    :param m: model
    :param in_: list of layers or tensors
    :param b: batch size to resize the model to
    :return: list of layers
    """
    # core module has multiple overloads for this:
    #  1. forward(m, in_)  where in_ is a list of layers
    #  2. forward(m, in_)  where in_ is a list of tensors
    #  3. forward(m)
    #  4. forward(m, b)
    if in_ is None:
        if b is None:
            return _eddl.forward(m)
        return _eddl.forward(m, b)
    return _eddl.forward(m, in_)


def zeroGrads(m):
    """\
    Set model gradients to zero.

    :param net: model
    :return: None
    """
    return _eddl.zeroGrads(m)


def backward(m, target=None):
    """\
    Calculate the gradient by passing its argument (1x1 unit tensor by
    default) through the backward graph.

    :param m: model or loss (if it's a loss then target must not be provided)
    :param target: list of tensors
    :return: None
    """
    # core module has multiple overloads for this:
    #  1. backward(m, target)  where m is a Model and target a list of Tensor
    #  2. backward(net)  where net is a Model
    #  3. backward(l)  where l is a NetLoss
    if target is None:
        return _eddl.backward(m)
    return _eddl.backward(m, target)


def optimize(l):
    return _eddl.optimize(l)


def update(m):
    """\
    Update the model weights.

    :param m: Model
    :return: None
    """
    return _eddl.update(m)


def print_loss(m, batch):
    """\
    Print model loss at the given batch.

    :param m: model
    :param batch: batch number
    :return: None
    """
    return _eddl.print_loss(m, batch)


def get_losses(m):
    """\
    Get model losses.

    :param m: model
    :return: list of float
    """
    return _eddl.get_losses(m)


def get_metrics(m):
    """\
    Get model metrics.

    :param m: model
    :return: list of float
    """
    return _eddl.get_metrics(m)


# = Model constraints =

def clamp(m, min, max):
    """\
    Perform model parameter clamping between min and max.

    :param m: model
    :param min: minimum value
    :param max: maximum value
    :return: None
    """
    return _eddl.clamp(m, min, max)


# = Loss and metrics methods =

def compute_loss(L):
    """\
    Compute the loss of the associated model.

    :param L: loss object
    :return: computed loss
    """
    return _eddl.compute_loss(L)


def compute_metric(L):
    """\
    Compute the loss of the associated model (alias for ``compute_loss``).

    :param L: loss object
    :return: computed loss
    """
    return _eddl.compute_metric(L)


def getLoss(type_):
    """\
    Get loss by name.

    :param type_: loss name
    :return: loss
    """
    return _eddl.getLoss(type_)


def newloss(f, in_, name):
    """\
    Create a new loss.

    ``f`` can be a Layer -> Layer function (and ``in_`` must be a Layer) or a
    [Layer] -> Layer function (and ``in_`` must be a [Layer]).

    :param f: loss function
    :param in_: loss input
    :param name: loss name
    :return: loss
    """
    # core module has multiple overloads for this:
    #  1. newloss(f, in_, name)  where:
    #    f is a function [Layer] -> Layer
    #    in is a [Layer]
    #  2. newloss(f, in_, name)  where:
    #    f is a function Layer -> Layer
    #    in is a Layer
    return _eddl.newloss(f, in_, name)


def getMetric(type_):
    """\
    Get Metric by name.

    :param type_: Metric name
    :return: metric
    """
    return _eddl.getMetric(type_)


# def newmetric(f, in_, name)  # TODO


def detach(l):
    """\
    Set a layer as detached, excluding it from gradient computation.

    :param l: layer or list of layers to detach
    :return: detached layer(s)
    """
    # core module has multiple overloads for this:
    #  1. detach(l)  where l is a Layer and the return value is a Layer
    #  2. detach(l)  where l is a [Layer] and the return value is a [Layer]
    return _eddl.detach(l)


# == LAYERS ==

# = Core layers =

def Activation(parent, activation, params=[], name=""):
    """\
    Apply an activation function to the given layer.

    :param parent: parent layer
    :param activation: name of the activation function
    :param params: list of floats (parameters of the activation function)
    :param name: name of the output layer
    :return: Activation layer
    """
    return _eddl.Activation(parent, activation, params, name)


def Softmax(parent, axis=1, name=""):
    """\
    Apply a softmax activation function to the given layer.

    :param parent: parent layer
    :param axis: dimension on which to operate
    :param name: name of the output layer
    :return: softmax Activation layer
    """
    return _eddl.Softmax(parent, axis, name)


def Sigmoid(parent, name=""):
    """\
    Apply a sigmoid activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: sigmoid Activation layer
    """
    return _eddl.Sigmoid(parent, name)


def HardSigmoid(parent, name=""):
    """\
    Apply a hard sigmoid activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: hard sigmoid Activation layer
    """
    return _eddl.HardSigmoid(parent, name)


def ReLu(parent, name=""):
    """\
    Apply a rectified linear unit activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: ReLu Activation layer
    """
    return _eddl.ReLu(parent, name)


def ThresholdedReLu(parent, alpha=1.0, name=""):
    """\
    Apply the thresholded version of a rectified linear unit activation
    function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: thresholded ReLu Activation layer
    """
    return _eddl.ThresholdedReLu(parent, alpha, name)


def LeakyReLu(parent, alpha=0.01, name=""):
    """\
    Apply the leaky version of a rectified linear unit activation
    function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: leaky ReLu Activation layer
    """
    return _eddl.LeakyReLu(parent, alpha, name)


def Elu(parent, alpha=1.0, name=""):
    """\
    Apply the exponential linear unit activation function to the given layer.

    :param parent: parent layer
    :param alpha: ELU coefficient
    :param name: name of the output layer
    :return: ELU Activation layer
    """
    return _eddl.Elu(parent, alpha, name)


def Selu(parent, name=""):
    """\
    Apply the scaled exponential linear unit activation function to the given
    layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: SELU Activation layer
    """
    return _eddl.Selu(parent, name)


def Exponential(parent, name=""):
    """\
    Apply the exponential (base e) activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: exponential Activation layer
    """
    return _eddl.Exponential(parent, name)


def Softplus(parent, name=""):
    """\
    Apply the softplus activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: softplus Activation layer
    """
    return _eddl.Softplus(parent, name)


def Softsign(parent, name=""):
    """\
    Apply the softsign activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: softsign Activation layer
    """
    return _eddl.Softsign(parent, name)


def Linear(parent, alpha=1.0, name=""):
    """\
    Apply the linear activation function to the given layer.

    :param parent: parent layer
    :param alpha: linear coefficient
    :param name: name of the output layer
    :return: linear Activation layer
    """
    return _eddl.Linear(parent, alpha, name)


def Tanh(parent, name=""):
    """\
    Apply the hyperbolic tangent activation function to the given layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: hyperbolic tangent Activation layer
    """
    return _eddl.Tanh(parent, name)


def Conv(parent, filters, kernel_size, strides=[1, 1], padding="same",
         use_bias=True, groups=1, dilation_rate=[1, 1], name=""):
    """\
    2D convolution layer.

    :param parent: parent layer
    :param filters: dimensionality of the output space (i.e., the number of
      output filters in the convolution)
    :param kernel_size: list of 2 integers, specifying the height and width of
      the 2D convolution window.
    :param strides: list of 2 integers, specifying the strides of the
      convolution along the height and width
    :param padding: one of "none", "valid" or "same"
    :param use_bias: whether the layer uses a bias vector
    :param groups: number of blocked connections from input to output channels
    :param dilation_rate: list of 2 integers, specifying the dilation rate
      to use for dilated convolution
    :param name: name of the output layer
    :return: Convolution layer
    """
    return _eddl.Conv(parent, filters, kernel_size, strides, padding,
                      use_bias, groups, dilation_rate, name)


def Conv1D(parent, filters, kernel_size, strides=[1], padding="same",
           use_bias=True, groups=1, dilation_rate=[1], name=""):
    """\
    1D convolution layer.

    :param parent: parent layer
    :param filters: dimensionality of the output space (i.e., the number of
      output filters in the convolution)
    :param kernel_size: list of 1 integer.
    :param strides: list of 1 integers
    :param padding: one of "none", "valid" or "same"
    :param use_bias: whether the layer uses a bias vector
    :param groups: number of blocked connections from input to output channels
    :param dilation_rate: list of 1 integer, specifying the dilation rate
      to use for dilated convolution
    :param name: name of the output layer
    :return: Convolution layer
    """
    return _eddl.Conv1D(parent, filters, kernel_size, strides, padding,
                        use_bias, groups, dilation_rate, name)


def PointwiseConv(parent, filters, strides=[1, 1], use_bias=True, groups=1,
                  dilation_rate=[1, 1], name=""):
    """\
    Pointwise convolution layer.

    :param parent: parent layer
    :param filters: dimensionality of the output space (i.e., the number of
      output filters in the convolution)
    :param strides: list of 2 integers, specifying the strides of the
      convolution along the height and width
    :param use_bias: whether the layer uses a bias vector
    :param groups: number of blocked connections from input to output channels
    :param dilation_rate: list of 2 integers, specifying the dilation rate
      to use for dilated convolution
    :param name: name of the output layer
    :return: Convolution layer
    """
    return _eddl.PointwiseConv(parent, filters, strides, use_bias, groups,
                               dilation_rate, name)


def Dense(parent, ndim, use_bias=True, name=""):
    """\
    Regular densely-connected layer.

    :param parent: parent layer
    :param ndim: dimensionality of the output space
    :param use_bias: whether the layer uses a bias vector
    :param name: name of the output layer
    :return: Dense layer
    """
    return _eddl.Dense(parent, ndim, use_bias, name)


def Dropout(parent, rate, iw=True, name=""):
    """\
    Apply dropout to a layer.

    The dropout consists of randomly setting a fraction of input units to
    0 at each update during training time, which helps prevent overfitting.

    :param parent: parent layer
    :param rate: fraction of input units to drop (between 0 and 1)
    :param iw: whether to perform weighting in inference
    :param name: name of the output layer
    :return: Dropout layer
    """
    return _eddl.Dropout(parent, rate, iw, name)


def Input(shape, name=""):
    """\
    Create a layer that can be used as input to a model.

    :param shape: list of dimensions, not including the batch size. For
      instance, shape=[32] indicates that the expected input will be batches
      of 32-dimensional vectors
    :param name: name of the output layer
    :return: Input layer
    """
    return _eddl.Input(shape, name)


def UpSampling(parent, size, interpolation="nearest", name=""):
    """\
    Upsampling layer.

    Similar to the scale transformation. The only difference is that
    upsampling repeats its rows/columns n times, while scaling uses a
    proportion.

    :param parent: parent layer
    :param size: list of 2 integers (upsampling factors for rows and columns)
    :param interpolation: "nearest" or "bilinear"
    :param name: name of the output layer
    :return: UpSampling layer
    """
    return _eddl.UpSampling(parent, size, interpolation, name)


def Reshape(parent, shape, name=""):
    """\
    Reshape an output to the given shape.

    :param parent: parent layer
    :param shape: target shape as a list of integers, not including the batch
      axis
    :param name: name of the output layer
    :return: Reshape layer
    """
    return _eddl.Reshape(parent, shape, name)


def Flatten(parent, name=""):
    """\
    Flatten the input. Does not affect the batch size.

    Equivalent to a :func:`.Reshape` with a shape of ``[-1]``.

    :param parent: parent layer
    :param name: name of the output layer
    :return: a Reshape layer
    """
    return _eddl.Flatten(parent, name)


def ConvT(parent, filters, kernel_size, output_padding, padding="same",
          dilation_rate=[1, 1], strides=[1, 1], use_bias=True, name=""):
    """\
    Transposed convolution layer (sometimes called deconvolution).

    The need for transposed convolutions generally arises from the desire to
    use a transformation going in the opposite direction of a normal
    convolution, i.e., from something that has the shape of the output of some
    convolution to something that has the shape of its input while maintaining
    a connectivity pattern that is compatible with said convolution.

    :param parent: parent layer
    :param filters: dimensionality of the output space (i.e., the number of
      output filters in the convolution)
    :param kernel_size: the height and width of the 2D convolution window
    :param output_padding: the amount of padding along the height and width of
      the output tensor. The amount of output padding along a given dimension
      must be lower than the stride along the same dimension
    :param padding: one of "valid" or "same"
    :param dilation_rate: the dilation rate to use for dilated convolution.
      Spacing between kernel elements
    :param strides: the strides of the convolution along the height and width
    :param use_bias: whether the layer uses a bias vector
    :param name: name of the output layer
    :return: ConvT layer

    """
    return _eddl.ConvT(parent, filters, kernel_size, output_padding, padding,
                       dilation_rate, strides, use_bias, name)


def Embedding(parent, vocsize, length, output_dim, mask_zeros=False, name=""):
    """\
    Turn positive integers (indexes) into dense vectors of fixed size. e.g.,
    ``[[4], [20]] -> [[0.25, 0.1], [0.6, -0.2]]``.

    :param parent: parent layer
    :param vocsize: size of the vocabulary, i.e., maximum integer index + 1
    :param length: length of the sequence, to connect to Dense layers
      (non Recurrent)
    :param output_dim: dimension of the dense embedding
    :param name: name of the output layer
    :return: Embedding layer
    """
    return _eddl.Embedding(
        parent, vocsize, length, output_dim, mask_zeros, name
    )


def Transpose(parent, name=""):
    """\
    Transpose a Layer.

    :param parent: parent layer
    :param name: name of the output layer
    :return: the transposed layer
    """
    return _eddl.Transpose(parent, name)


# = Transformation layers =

def Crop(parent, from_coords, to_coords, reshape=True, constant=0.0, name=""):
    """\
    Crop the given image layer at ``[(top, left), (bottom, right)]``.

    :param parent: parent layer
    :param from_coords: [top, left] coordinates
    :param to_coords: [bottom, right] coordinates
    :param reshape: if True, the output shape will be new_shape (classical
      scale; recommended). If False, the output shape will be the input shape
      (scale < 100%: scale + padding; scale > 100%: crop + scale)
    :param constant: erasing value
    :param name: name of the output layer
    :return: Crop layer
    """
    return _eddl.Crop(parent, from_coords, to_coords, reshape, constant, name)


def CenteredCrop(parent, size, reshape=True, constant=0.0, name=""):
    """\
    Crop the given image layer at the center with size ``[width, height]``.

    :param parent: parent layer
    :param size: [height, width]
    :param reshape: If True, the output shape will be new_shape (classical
      scale; recommended). If False, the output shape will be the input shape
      (scale < 100%: scale + padding; scale > 100%: crop + scale)
    :param constant: erasing value
    :param name: name of the output layer
    :return: a Crop layer
    """
    return _eddl.CenteredCrop(parent, size, reshape, constant, name)


def CropScale(parent, from_coords, to_coords, da_mode="constant", constant=0.0,
              name=""):
    """\
    Crop the given image layer at ``[(top, left), (bottom, right)]`` and scale
    it to the parent size.

    :param parent: parent layer
    :param from_coords: [top, left] coordinates
    :param to_coords: [bottom, right] coordinates
    :param da_mode: one of "nearest", "constant"
    :param constant: fill value for the area outside the rotated image, used
      for all channels
    :param name: name of the output layer
    :return: CropScale layer
    """
    return _eddl.CropScale(parent, from_coords, to_coords, da_mode, constant,
                           name)


def Cutout(parent, from_coords, to_coords, constant=0.0, name=""):
    """\
    Select a rectangle in an image layer at ``[(top, left), (bottom, right)]``
    and erase its pixels using a constant value.

    :param parent: parent layer
    :param from_coords: [top, left] coordinates
    :param to_coords: [bottom, right] coordinates
    :param constant: erasing value
    :param name: name of the output layer
    :return: Cutout layer
    """
    return _eddl.Cutout(parent, from_coords, to_coords, constant, name)


def Flip(parent, axis=0, name=""):
    """\
    Flip an image layer at the given axis.

    :param parent: parent layer
    :param axis: flip axis
    :param name: name of the output layer
    :return: Flip layer
    """
    return _eddl.Flip(parent, axis, name)


def HorizontalFlip(parent, name=""):
    """\
    Flip an image layer horizontally.

    :param parent: parent layer
    :param name: name of the output layer
    :return: a Flip layer
    """
    return _eddl.HorizontalFlip(parent, name)


def Rotate(parent, angle, offset_center=[0, 0], da_mode="original",
           constant=0.0, name=""):
    """\
    Rotate an image layer by the given angle, counterclockwise.

    :param parent: parent layer
    :param angle: rotation angle in degrees
    :param offset_center: center of rotation
    :param da_mode: one of "nearest", "constant"
    :param constant: fill value for the area outside the rotated image, used
      for all channels
    :return: Rotate layer
    """
    return _eddl.Rotate(parent, angle, offset_center, da_mode, constant, name)


def Scale(parent, new_shape, reshape=True, da_mode="constant", constant=0.0,
          name=""):
    """\
    Resize an image layer to the given size as ``[height, width]``.

    :param parent: parent layer
    :param new_shape: new shape
    :param reshape: if True, the output shape will be new_shape (classical
      scale; recommended). If False, the output shape will be the input shape
      (scale < 100%: scale + padding; scale > 100%: crop + scale)
    :param da_mode: one of "nearest", "constant"
    :param constant: fill value for the area outside the resized image, used
      for all channels
    :return: Scale layer
    """
    return _eddl.Scale(parent, new_shape, reshape, da_mode, constant, name)


def Shift(parent, shift, da_mode="nearest", constant=0.0, name=""):
    """\
    Shift the input image.

    :param parent: parent layer
    :param shift: list of maximum absolute fraction for the horizontal and
      vertical translations
    :param da_mode: one of "nearest", "constant"
    :param constant: fill value for the area outside the resized image, used
      for all channels
    :return: Shift layer
    """
    return _eddl.Shift(parent, shift, da_mode, constant, name)


def VerticalFlip(parent, name=""):
    """\
    Flip an image layer vertically.

    :param parent: parent layer
    :param name: name of the output layer
    :return: a Flip layer
    """
    return _eddl.VerticalFlip(parent, name)


# = Data augmentation layers =

def RandomCrop(parent, new_shape, name=""):
    """\
    Crop an image layer at a random location with size ``[height, width]``.

    :param parent: parent layer
    :param new_shape: [height, width] size
    :param name: name of the output layer
    :return: CropRandom layer
    """
    return _eddl.RandomCrop(parent, new_shape, name)


def RandomCropScale(parent, factor, da_mode="nearest", name=""):
    """\
    Crop an image layer randomly and scale it to the parent size.

    :param parent: parent layer
    :param factor: crop range factor
    :param da_mode: one of "nearest", "constant"
    :param name: name of the output layer
    :return: CropScaleRandom layer
    """
    return _eddl.RandomCropScale(parent, factor, da_mode, name)


def RandomCutout(parent, factor_x, factor_y, constant=0.0, name=""):
    """\
    Randomly select a rectangle region in an image layer and erase its pixels.

    The random region is defined by the range ``[(min_x, max_x), (min_y,
    max_y)]`` (relative values).

    :param parent: parent layer
    :param factor_x: list of factors for horizontal size
    :param factor_y: list of factors for vertical size
    :param constant: erasing value
    :param name: name of the output layer
    :return: CutoutRandom layer
    """
    return _eddl.RandomCutout(parent, factor_x, factor_y, constant, name)


def RandomFlip(parent, axis, name=""):
    """\
    Flip an image layer at the given axis randomly.

    :param parent: parent layer
    :param axis: flip axis
    :param name: name of the output layer
    :return: FlipRandom layer
    """
    return _eddl.RandomFlip(parent, axis, name)


def RandomHorizontalFlip(parent, name=""):
    """\
    Flip an image layer horizontally, randomly.

    :param parent: parent layer
    :param name: name of the output layer
    :return: a FlipRandom layer
    """
    return _eddl.RandomHorizontalFlip(parent, name)


def RandomRotation(parent, factor, offset_center=[0, 0], da_mode="original",
                   constant=0.0, name=""):
    """\
    Rotate an image layer randomly.

    :param parent: parent layer
    :param factor: angle range in degrees (counterclockwise)
    :param offset_center: center of rotation
    :param da_mode: one of "original"
    :param constant: fill value for the area outside the rotated image, used
      for all channels.
    :param name: name of the output layer
    :return: RotateRandom layer
    """
    return _eddl.RandomRotation(parent, factor, offset_center, da_mode,
                                constant, name)


def RandomScale(parent, factor, da_mode="nearest", constant=0.0, name=""):
    """\
    Resize an image layer randomly.

    :param parent: parent layer
    :param factor: list of resize factors for the new shape
    :param da_mode: One of "nearest"
    :param constant: fill value for the area outside the resized image,
      used for all channels
    :param name: name of the output layer
    :return: ScaleRandom layer
    """
    return _eddl.RandomScale(parent, factor, da_mode, constant, name)


def RandomShift(parent, factor_x, factor_y, da_mode="nearest", constant=0.0,
                name=""):
    """\
    Shift an image layer randomly.

    :param parent: parent layer
    :param factor_x: list of factors for horizontal translations
    :param factor_y: list of factors for vertical translations
    :param da_mode: one of "nearest", "constant"
    :param constant: fill value for the area outside the resized image, used
      for all channels
    :param name: name of the output layer
    :return: ShiftRandom layer
    """
    return _eddl.RandomShift(parent, factor_x, factor_y, da_mode, constant,
                             name)


def RandomVerticalFlip(parent, name=""):
    """\
    Flip an image layer vertically, randomly.

    :param parent: parent layer
    :param name: name of the output layer
    :return: a RandomFlip layer
    """
    return _eddl.RandomVerticalFlip(parent, name)


# = Merge layers =

def Add(layers, name=""):
    """\
    Add input layers.

    **NOTE:** this function can also be used to compute the sum of two layers
    or floats (which was previously done via the now deprecated function
    Sum). In this case, the function is called as ``Add(l1, l2)``, where each
    of the two parameters can be a layer or a float (one of them must be a
    layer).

    :param layers: list of layers, all of the same shape
    :param name: name of the output layer
    :return: Add layer
    """
    return _eddl.Add(layers, name)


def Average(layers, name=""):
    """\
    Compute the average of a list of input layers.

    :param layers: list of layers, all of the same shape
    :param name: name of the output layer
    :return: Average layer
    """
    return _eddl.Average(layers, name)


def Concat(layers, axis=1, name=""):
    """\
    Concatenate input layers.

    :param layers: list of layers
    :param axis: axis along which to concatenate
    :param name: name of the output layer
    :return: Concat layer
    """
    return _eddl.Concat(layers, axis, name)


def MatMul(layers, name=""):
    return _eddl.MatMul(layers, name)


def Maximum(layers, name=""):
    """\
    Compute the maximum (element-wise) of a list of input layers.

    :param layers: list of layers, all of the same shape
    :param name: name of the output layer
    :return: Maximum layer
    """
    return _eddl.Maximum(layers, name)


def Minimum(layers, name=""):
    """\
    Compute the minimum (element-wise) of a list of input layers.

    :param layers: list of layers, all of the same shape
    :param name: name of the output layer
    :return: Minimum layer
    """
    return _eddl.Minimum(layers, name)


def Subtract(layers, name=""):
    """\
    Subtract two input layers.

    :param layers: list of two layers with the same shape
    :param name: name of the output layer
    :return: Substract layer
    """
    return _eddl.Subtract(layers, name)


# = Noise layers =

def GaussianNoise(parent, stddev, name=""):
    """\
    Apply additive zero-centered Gaussian noise.

    This is useful to mitigate overfitting (can be considered a form of
    random data augmentation). Gaussian Noise (GS) is a natural choice as
    corruption process for real valued inputs. Being a regularization layer,
    it is only active at training time.

    :param parent: parent layer
    :param stddev: standard deviation of the noise distribution
    :param name: name of the output layer
    :return: GaussianNoise layer
    """
    return _eddl.GaussianNoise(parent, stddev, name)


# = Normalization =

def BatchNormalization(parent, affine, momentum=0.9, epsilon=0.00001, name=""):
    """\
    Batch normalization layer.

    Normalize the activations of the input layer at each batch, i.e., apply a
    transformation that maintains the mean activation close to 0 and the
    activation standard deviation close to 1. See:
    https://arxiv.org/abs/1502.03167

    :param parent: parent layer
    :param affine: if True, this module has learnable affine parameters
    :param momentum: momentum for the moving mean and the moving variance
    :param epsilon: small float added to variance to avoid dividing by zero
    :param name: name of the output layer
    :return: BatchNorm layer
    """
    return _eddl.BatchNormalization(parent, affine, momentum, epsilon, name)


def LayerNormalization(parent, affine=True, epsilon=0.00001, name=""):
    """\
    Layer normalization layer.

    See: https://arxiv.org/abs/1607.06450.

    :param parent: parent layer
    :param affine: if True, this module has learnable affine parameters
    :param momentum: momentum for the moving mean and the moving variance
    :param epsilon: value added to the denominator for numerical stability
    :param name: name of the output layer
    :return: LayerNorm layer
    """
    return _eddl.LayerNormalization(parent, affine, epsilon, name)


def GroupNormalization(parent, groups, epsilon=0.001, affine=True, name=""):
    """\
    Group normalization layer.

    Divide the channels into groups and compute within each group the mean
    and variance for normalization. The computation is independent of batch
    sizes. See: https://arxiv.org/abs/1803.08494.

    :param parent: parent layer
    :param groups: number of groups in which the channels will be divided
    :param momentum: momentum for the moving mean and the moving variance
    :param epsilon: value added to the denominator for numerical stability
    :param affine: if True, this module has learnable affine parameters
    :param name: name of the output layer
    :return: GroupNorm layer
    """
    return _eddl.GroupNormalization(parent, groups, epsilon, affine, name)


def Norm(parent, epsilon=0.001, name=""):
    return _eddl.Norm(parent, epsilon, name)


def NormMax(parent, epsilon=0.001, name=""):
    return _eddl.NormMax(parent, epsilon, name)


def NormMinMax(parent, epsilon=0.001, name=""):
    return _eddl.NormMinMax(parent, epsilon, name)


# = Operator layers =

def Abs(l):
    """\
    Compute the element-wise absolute value of the given input layer.

    :param l: parent layer
    :return: Abs layer
    """
    return _eddl.Abs(l)


def Sub(l1, l2):
    """\
    Compute the difference between two layers or floats.

    :param l1: a layer or float
    :param l2: a layer or float
    :return: Sub layer
    """
    # l1, l2 can be either layers or floats
    return _eddl.Sub(l1, l2)


def Diff(l1, l2):
    """\
    Compute the difference between two layers or floats.

    Deprecated alias for Sub.

    :param l1: a layer or float
    :param l2: a layer or float
    :return: Diff layer
    """
    # l1, l2 can be either layers or floats
    return _eddl.Diff(l1, l2)


def Div(l1, l2):
    """\
    Compute the element-wise division of two layers or floats.

    :param l1: a layer or float
    :param l2: a layer or float
    :return: Div layer
    """
    # l1, l2 can be either layers or floats
    return _eddl.Div(l1, l2)


def Exp(l):
    """\
    Compute the element-wise exponential of the input layer.

    :param l: parent layer
    :return: Exp layer
    """
    return _eddl.Exp(l)


def Log(l):
    """\
    Compute the logarithm of the input layer.

    :param l: parent layer
    :return: Log layer
    """
    return _eddl.Log(l)


def Log2(l):
    """\
    Compute the base 2 logarithm of the input layer.

    :param l: parent layer
    :return: Log2 layer
    """
    return _eddl.Log2(l)


def Log10(l):
    """\
    Compute the base 10 logarithm of the input layer.

    :param l: parent layer
    :return: Log10 layer
    """
    return _eddl.Log10(l)


def Mult(l1, l2):
    """\
    Compute the element-wise multiplication of two layers or floats.

    :param l1: a layer or float
    :param l2: a layer or float
    :return: Mult layer
    """
    # l1, l2 can be either layers or floats
    return _eddl.Mult(l1, l2)


def Pow(l1, l2):
    """\
    Compute the element-wise power of two layers, or a layer and a float.

    :param l1: a layer
    :param l2: a layer or float
    :return: Pow layer
    """
    # l2 can be either a layer or a float
    return _eddl.Pow(l1, l2)


def Sqrt(l):
    """\
    Compute the square root of a layer.

    :param l: parent layer
    :return: Sqrt layer
    """
    return _eddl.Sqrt(l)


def Sum(l1, l2):
    """\
    Compute the sum of two layers or floats.

    Deprecated alias for Add (in the add-two-layers-or-floats version).

    :param l1: a layer or float
    :param l2: a layer or float
    :return: Sum layer
    """
    # l1, l2 can be either layers or floats
    return _eddl.Sum(l1, l2)


def Select(l, indices, name=""):
    """\
    Create a new layer which indexes the input layer using the entries in
    ``indices``.

    :param l: parent layer
    :param indices: list of indices to be selected
    :param name: name of the output layer
    :return: Select layer
    """
    return _eddl.Select(l, indices, name)


def Permute(l, dims, name=""):
    """\
    Permute the dimensions of the input according to a given pattern.

    :param l: parent layer
    :param dims: permutation pattern, does not include the samples dimension
    :param name: name of the output layer
    :return: Permute layer
    """
    return _eddl.Permute(l, dims, name)


# = Reduction layers =

def ReduceMean(l, axis, keepdims=False):
    return _eddl.ReduceMean(l, axis, keepdims)


def ReduceVar(l, axis, keepdims=False):
    return _eddl.ReduceVar(l, axis, keepdims)


def ReduceSum(l, axis, keepdims=False):
    return _eddl.ReduceSum(l, axis, keepdims)


def ReduceMax(l, axis, keepdims=False):
    return _eddl.ReduceMax(l, axis, keepdims)


def ReduceMin(l, axis, keepdims=False):
    return _eddl.ReduceMin(l, axis, keepdims)


def ReduceArgMax(l, axis, keepdims=False):
    return _eddl.ReduceArgMax(l, axis, keepdims)


# = Generator layers =

def GaussGenerator(mean, stdev, size):
    return _eddl.GaussGenerator(mean, stdev, size)


def UniformGenerator(low, high, size):
    return _eddl.UniformGenerator(low, high, size)


# = Pooling layers =

def AveragePool(parent, pool_size=[2, 2], strides=[2, 2], padding="none",
                name=""):
    """\
    Perform average pooling.

    :param parent: parent layer
    :param pool_size: size of the average pooling windows
    :param strides: factors by which to downscale
    :param padding: one of "none", "valid" or "same"
    :param name: name of the output layer
    :return: AveragePool layer
    """
    return _eddl.AveragePool(parent, pool_size, strides, padding, name)


def GlobalMaxPool(parent, name=""):
    """\
    Perform global max pooling.

    :param parent: parent layer
    :param name: name of the output layer
    :return: a MaxPool layer
    """
    return _eddl.GlobalMaxPool(parent, name)


def GlobalAveragePool(parent, name=""):
    """\
    Perform global average pooling.

    :param parent: parent layer
    :param name: name of the output layer
    :return: an AveragePool layer
    """
    return _eddl.GlobalAveragePool(parent, name)


def MaxPool(parent, pool_size=[2, 2], strides=[2, 2], padding="none", name=""):
    """\
    Perform Max pooling.

    :param parent: parent layer
    :param pool_size: size of the max pooling windows
    :param strides: factor by which to downscale
    :param padding: one of "none", "valid" or "same"
    :param name: name of the output layer
    :return: MaxPool layer
    """
    return _eddl.MaxPool(parent, pool_size, strides, padding, name)


def MaxPool1D(parent, pool_size=[2], strides=[2], padding="none", name=""):
    """\
    Perform 1D Max pooling.

    :param parent: parent layer
    :param pool_size: size of the max pooling windows
    :param strides: factor by which to downscale
    :param padding: one of "none", "valid" or "same"
    :param name: name of the output layer
    :return: MaxPool1D layer
    """
    return _eddl.MaxPool1D(parent, pool_size, strides, padding, name)


# = Recurrent layers =

def RNN(parent, units, activation="tanh", use_bias=True, bidirectional=False,
        name=""):
    """\
    Fully-connected RNN where the output is to be fed back to input.

    :param parent: parent layer
    :param units: dimensionality of the output space.
    :param activation: activation
    :param use_bias: whether the layer uses a bias vector
    :param bidirectional: whether the RNN is bidirectional
    :param name: name of the output layer
    :return: RNN layer
    """
    return _eddl.RNN(parent, units, activation, use_bias, bidirectional, name)


def LSTM(parent, units, mask_zeros=False, bidirectional=False, name=""):
    """\
    Long Short-Term Memory layer - Hochreiter 1997.

    :param parent: parent layer
    :param units: dimensionality of the output space.
    :param mask_zeros: boolean
    :param bidirectional: whether the net is bidirectional or not
    :param name: name of the output layer
    :return: LSTM layer
    """
    return _eddl.LSTM(parent, units, mask_zeros, bidirectional, name)


def Decoder(l, ld, op="concat"):
    return _eddl.Decoder(l, ld, op)


# = Layers methods =

def setTrainable(net, lanme, val):
    return _eddl.setTrainable(net, lanme, val)


def getOut(net):
    return _eddl.getOut(net)


# = Manage tensors inside layers =

def getOutput(l1):
    return _eddl.getOutput(l1)


def getDelta(l1):
    return _eddl.getDelta(l1)


def getParam(l1, p):
    return _eddl.getParam(l1, p)


def getGradient(l1, p):
    return _eddl.getGradient(l1, p)


def getParams(l1):
    return _eddl.getParams(l1)


def getGradients(l1):
    return _eddl.getGradients(l1)


def copyOutput(l1, l2):
    return _eddl.copyOutput(l1, l2)


def copyDelta(l1, l2):
    return _eddl.copyDelta(l1, l2)


def copyParam(l1, l2, p):
    return _eddl.copyParam(l1, l2, p)


def copyGradient(l1, l2, p):
    return _eddl.copyGradient(l1, l2, p)


# == INITIALIZERS ==

def GlorotNormal(l, seed=1234):
    """\
    Glorot normal initializer, also called Xavier normal initializer.

    It draws samples from a truncated normal distribution centered on 0 with
    ``stddev = sqrt(2 / (fan_in + fan_out))`` where fan_in is the number of
    input units in the weight tensor and fan_out is the number of output units
    in the weight tensor.

    :param l: parent layer to initialize
    :param seed: used to seed the random generator
    :return: GlorotNormal layer
    """
    return _eddl.GlorotNormal(l, seed)


def GlorotUniform(l, seed=1234):
    """\
    Glorot uniform initializer, also called Xavier uniform initializer.

    It draws samples from a uniform distribution within [-limit, limit] where
    limit is ``sqrt(6 / (fan_in + fan_out))``, where fan_in is the number of
    input units in the weight tensor and fan_out is the number of output units
    in the weight tensor.

    :param l: parent layer to initialize
    :param seed: used to seed the random generator
    :return: GlorotUniform layer
    """
    return _eddl.GlorotUniform(l, seed)


def HeNormal(l, seed=1234):
    """\
    He normal initializer.

    It draws samples from a truncated normal distribution centered on 0 with
    stddev = sqrt(2 / (fan_in)) where fan_in is the number of input units in
    the weight tensor.

    :param l: parent layer to initialize
    :param seed: used to seed the random generator
    :return: HeNormal layer
    """
    return _eddl.HeNormal(l, seed)


def HeUniform(l, seed=1234):
    """\
    He uniform initializer.

    It draws samples from a uniform distribution within [-limit, limit] where
    limit is sqrt(6 / (fan_in )) where fan_in is the number of input units in
    the weight tensor.

    :param l: parent layer to initialize
    :param seed: used to seed the random generator
    :return: HeUniform layer
    """
    return _eddl.HeUniform(l, seed)


def RandomNormal(l, m=0.0, s=0.1, seed=1234):
    """\
    Random normal initializer.

    :param l: parent layer to initialize
    :param m: mean of the normal distribution to draw samples
    :param s: standard deviation of the normal distribution to draw samples
    :param seed: used to seed the random generator
    :return: RandomNormal layer
    """
    return _eddl.RandomNormal(l, m, s, seed)


def RandomUniform(l, min=0.0, max=0.1, seed=1234):
    """\
    Random uniform initializer.

    :param l: parent layer to initialize
    :param min: min of the distribution
    :param max: max of the distribution
    :param seed: used to seed the random generator
    :return: RandomUniform layer
    """
    return _eddl.RandomUniform(l, min, max, seed)


def Constant(l, v=0.1):
    """\
    Initializer that generates tensors initialized to a constant value.

    :param l: parent layer to initialize
    :param v: value of the generator
    :return: Constant layer
    """
    return _eddl.Constant(l, v)


# == REGULARIZERS ==

def L2(l, l2):
    """\
    Regularizer for L2 regularization.

    :param l: parent layer to regularize
    :param l2: L2 regularization factor
    :return: the input layer, regularized
    """
    return _eddl.L2(l, l2)


def L1(l, l1):
    """\
    Regularizer for L1 regularization.

    :param l: parent layer to regularize
    :param l1: L1 regularization factor
    :return: the input layer, regularized
    """
    return _eddl.L1(l, l1)


def L1L2(l, l1, l2):
    """\
    Regularizer for L1 and L2 regularization.

    :param l: parent layer to regularize
    :param l1: L1 regularization factor
    :param l2: L2 regularization factor
    :return: the input layer, regularized
    """
    return _eddl.L1L2(l, l1, l2)


# == DATASETS ==

def exist(name):
    return _eddl.exist(name)


def download_mnist():
    """\
    Download the MNIST dataset.

    See: http://yann.lecun.com/exdb/mnist/
    """
    return _eddl.download_mnist()


def download_cifar10():
    """\
    Download the CIFAR-10 Dataset.

    See: https://www.cs.toronto.edu/~kriz/cifar.html
    """
    return _eddl.download_cifar10()


def download_drive():
    """\
    Download the DRIVE Dataset.

    See: https://drive.grand-challenge.org/
    """
    return _eddl.download_drive()


def download_imdb_2000():
    """\
    Download the IMDB Dataset, 2000 most frequent words.

    See: https://ai.stanford.edu/~amaas/data/sentiment/
    """
    return _eddl.download_imdb_2000()


def download_eutrans():
    """\
    Download the EuTrans Dataset.
    """
    return _eddl.download_eutrans()


def download_flickr():
    """\
    Download the Flickr Dataset (small partition).
    """
    return _eddl.download_flickr()


# == ONNX ==

def save_net_to_onnx_file(net, path):
    return _eddl.save_net_to_onnx_file(net, path)


def import_net_from_onnx_file(path):
    return _eddl.import_net_from_onnx_file(path)


def serialize_net_to_onnx_string(net, gradients):
    return _eddl.serialize_net_to_onnx_string(net, gradients)


def import_net_from_onnx_string(model_string):
    return _eddl.import_net_from_onnx_string(model_string)



# Copyright (c) 2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
MNIST encoder-decoder with user-defined losses.
"""

import argparse
import sys

import pyeddl.eddl as eddl
from pyeddl.tensor import Tensor


# l2_loss
def mse_loss(inputs):
    diff = eddl.Sub(inputs[0], inputs[1])
    return eddl.Mult(diff, diff)


# Dice loss image-level
def dice_loss_img(inputs):
    num = eddl.Mult(2, eddl.ReduceSum(
        eddl.Mult(inputs[0], inputs[1]), [0, 1, 2]
    ))
    den = eddl.ReduceSum(eddl.Add(inputs[0], inputs[1]), [0, 1, 2])
    return eddl.Sub(1.0, eddl.Div(num, den))


# Dice loss pixel-level
def dice_loss_pixel(inputs):
    num = eddl.Mult(2, eddl.ReduceSum(eddl.Mult(inputs[0], inputs[1]), [0]))
    den = eddl.ReduceSum(eddl.Add(inputs[0], inputs[1]), [0])
    num = eddl.Add(num, 1)
    den = eddl.Add(den, 1)
    return eddl.Sub(1.0, eddl.Div(num, den))


MEM_CHOICES = ("low_mem", "mid_mem", "full_mem")


def main(args):
    eddl.download_mnist()

    in_ = eddl.Input([784])
    target = eddl.Reshape(in_, [1, 28, 28])
    layer = in_
    layer = eddl.Reshape(layer, [1, 28, 28])
    layer = eddl.ReLu(eddl.Conv(layer, 8, [3, 3]))
    layer = eddl.ReLu(eddl.Conv(layer, 16, [3, 3]))
    layer = eddl.ReLu(eddl.Conv(layer, 8, [3, 3]))
    out = eddl.Sigmoid(eddl.Conv(layer, 1, [3, 3]))
    net = eddl.Model([in_], [])

    eddl.build(
        net,
        eddl.adam(0.001),
        [],
        [],
        eddl.CS_GPU(mem=args.mem) if args.gpu else eddl.CS_CPU(mem=args.mem)
    )
    eddl.summary(net)

    x_train = Tensor.load("mnist_trX.bin")
    if args.small:
        x_train = x_train.select([":6000"])
    x_train.div_(255.0)

    mse = eddl.newloss(mse_loss, [out, target], "mse_loss")
    dicei = eddl.newloss(dice_loss_img, [out, target], "dice_loss_img")
    dicep = eddl.newloss(dice_loss_pixel, [out, target], "dice_loss_pixel")

    batch = Tensor([args.batch_size, 784])
    num_batches = x_train.shape[0] // args.batch_size
    for i in range(args.epochs):
        print("Epoch %d/%d (%d batches)" % (i + 1, args.epochs, num_batches))
        diceploss = 0.0
        diceiloss = 0.0
        mseloss = 0
        for j in range(num_batches):
            print("Batch %d " % j, end="", flush=True)
            eddl.next_batch([x_train], [batch])
            eddl.zeroGrads(net)
            eddl.forward(net, [batch])
            diceploss += eddl.compute_loss(dicep) / args.batch_size
            print("diceploss = %.6f " % (diceploss / (j + 1)), end="")
            diceiloss += eddl.compute_loss(dicei) / args.batch_size
            print("diceiloss = %.6f " % (diceiloss / (j + 1)), end="")
            mseloss += eddl.compute_loss(mse) / args.batch_size
            print("mseloss = %.6f\r" % (mseloss / (j + 1)), end="")
            eddl.optimize(dicep)
            eddl.update(net)
        print()
    print("All done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--epochs", type=int, metavar="INT", default=10)
    parser.add_argument("--batch-size", type=int, metavar="INT", default=1000)
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--small", action="store_true")
    parser.add_argument("--mem", metavar="|".join(MEM_CHOICES),
                        choices=MEM_CHOICES, default="low_mem")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Check usage of getOut and forward in a network with a non-empty lout.
"""

import argparse
import os
import sys

import pyeddl.eddl as eddl
from pyeddl.tensor import Tensor


MEM_CHOICES = ("low_mem", "mid_mem", "full_mem")


def main(args):
    if not os.path.isfile(args.input):
        raise RuntimeError("input file '%s' not found" % args.input)

    eddl.download_mnist()

    net = eddl.import_net_from_onnx_file(args.input)
    eddl.build(
        net,
        eddl.rmsprop(0.01),
        ["soft_cross_entropy"],
        ["categorical_accuracy"],
        eddl.CS_GPU(mem=args.mem) if args.gpu else eddl.CS_CPU(mem=args.mem),
        False  # do not initialize weights to random values
    )

    net.resize(args.batch_size)  # resize manually since we don't use "fit"
    eddl.summary(net)

    x_test = Tensor.load("mnist_tsX.bin")
    x_test.div_(255.0)

    sys.stderr.write("forward...\n")
    eddl.forward(net, [x_test])
    sys.stderr.write("forward done\n")

    sys.stderr.write("lout: %r\n" % (net.lout,))
    out = eddl.getOut(net)
    sys.stderr.write("getOut done\n")
    sys.stderr.write("out: %r\n" % (out,))
    print("All done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--batch-size", type=int, metavar="INT", default=1000)
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--input", metavar="STRING",
                        default="trained_model.onnx",
                        help="input path of the serialized model")
    parser.add_argument("--mem", metavar="|".join(MEM_CHOICES),
                        choices=MEM_CHOICES, default="low_mem")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Drive segmentation.

https://drive.grand-challenge.org/DRIVE
"""

import argparse
import sys

import pyeddl.eddl as eddl
from pyeddl.tensor import Tensor


USE_CONCAT = 1


def UNetWithPadding(layer):
    x = layer
    depth = 32

    x = eddl.LeakyReLu(eddl.Conv(x, depth, [3, 3], [1, 1], "same"))
    x = eddl.LeakyReLu(eddl.Conv(x, depth, [3, 3], [1, 1], "same"))
    x2 = eddl.MaxPool(x, [2, 2], [2, 2])
    x2 = eddl.LeakyReLu(eddl.Conv(x2, 2*depth, [3, 3], [1, 1], "same"))
    x2 = eddl.LeakyReLu(eddl.Conv(x2, 2*depth, [3, 3], [1, 1], "same"))
    x3 = eddl.MaxPool(x2, [2, 2], [2, 2])
    x3 = eddl.LeakyReLu(eddl.Conv(x3, 4*depth, [3, 3], [1, 1], "same"))
    x3 = eddl.LeakyReLu(eddl.Conv(x3, 4*depth, [3, 3], [1, 1], "same"))
    x4 = eddl.MaxPool(x3, [2, 2], [2, 2])
    x4 = eddl.LeakyReLu(eddl.Conv(x4, 8*depth, [3, 3], [1, 1], "same"))
    x4 = eddl.LeakyReLu(eddl.Conv(x4, 8*depth, [3, 3], [1, 1], "same"))
    x5 = eddl.MaxPool(x4, [2, 2], [2, 2])
    x5 = eddl.LeakyReLu(eddl.Conv(x5, 8*depth, [3, 3], [1, 1], "same"))
    x5 = eddl.LeakyReLu(eddl.Conv(x5, 8*depth, [3, 3], [1, 1], "same"))
    x5 = eddl.Conv(
        eddl.UpSampling(x5, [2, 2]), 8*depth, [2, 2], [1, 1], "same"
    )

    x4 = eddl.Concat([x4, x5]) if USE_CONCAT else eddl.Add([x4, x5])
    x4 = eddl.LeakyReLu(eddl.Conv(x4, 8*depth, [3, 3], [1, 1], "same"))
    x4 = eddl.LeakyReLu(eddl.Conv(x4, 8*depth, [3, 3], [1, 1], "same"))
    x4 = eddl.Conv(
        eddl.UpSampling(x4, [2, 2]), 4*depth, [2, 2], [1, 1], "same"
    )

    x3 = eddl.Concat([x3, x4]) if USE_CONCAT else eddl.Add([x3, x4])
    x3 = eddl.LeakyReLu(eddl.Conv(x3, 4*depth, [3, 3], [1, 1], "same"))
    x3 = eddl.LeakyReLu(eddl.Conv(x3, 4*depth, [3, 3], [1, 1], "same"))
    x3 = eddl.Conv(
        eddl.UpSampling(x3, [2, 2]), 2*depth, [2, 2], [1, 1], "same"
    )

    x2 = eddl.Concat([x2, x3]) if USE_CONCAT else eddl.Add([x2, x3])
    x2 = eddl.LeakyReLu(eddl.Conv(x2, 2*depth, [3, 3], [1, 1], "same"))
    x2 = eddl.LeakyReLu(eddl.Conv(x2, 2*depth, [3, 3], [1, 1], "same"))
    x2 = eddl.Conv(
        eddl.UpSampling(x2, [2, 2]), depth, [2, 2], [1, 1], "same"
    )

    x = eddl.Concat([x, x2]) if USE_CONCAT else eddl.Add([x, x2])
    x = eddl.LeakyReLu(eddl.Conv(x, depth, [3, 3], [1, 1], "same"))
    x = eddl.LeakyReLu(eddl.Conv(x, depth, [3, 3], [1, 1], "same"))
    x = eddl.Conv(x, 1, [1, 1])

    return x


MEM_CHOICES = ("low_mem", "mid_mem", "full_mem")


def main(args):
    eddl.download_drive()

    in_1 = eddl.Input([3, 584, 584])
    in_2 = eddl.Input([1, 584, 584])
    layer = eddl.Concat([in_1, in_2])

    layer = eddl.RandomCropScale(layer, [0.9, 1.0])
    layer = eddl.CenteredCrop(layer, [512, 512])
    img = eddl.Select(layer, ["0:3"])
    mask = eddl.Select(layer, ["3"])

    # DA net
    danet = eddl.Model([in_1, in_2], [])
    eddl.build(danet)
    if args.gpu:
        eddl.toGPU(danet, mem="low_mem")
    eddl.summary(danet)

    # SegNet
    in_ = eddl.Input([3, 512, 512])
    out = eddl.Sigmoid(UNetWithPadding(in_))
    segnet = eddl.Model([in_], [out])
    eddl.build(
        segnet,
        eddl.adam(0.00001),  # Optimizer
        ["mse"],  # Losses
        ["mse"],  # Metrics
        eddl.CS_GPU(mem=args.mem) if args.gpu else eddl.CS_CPU(mem=args.mem)
    )
    eddl.summary(segnet)

    print("Reading training data")
    # x_train_f = Tensor.fromarray(np.load("drive_trX.npy").astype(np.float32))
    x_train_f = Tensor.load("drive_trX.bin")
    x_train = x_train_f.permute([0, 3, 1, 2])
    x_train.info()
    x_train.div_(255.0)

    print("Reading test data")
    # y_train = Tensor.fromarray(np.load("drive_trY.npy").astype(np.float32))
    y_train = Tensor.load("drive_trY.bin")
    y_train.info()
    y_train.reshape_([20, 1, 584, 584])
    y_train.div_(255.0)

    xbatch = Tensor([args.batch_size, 3, 584, 584])
    ybatch = Tensor([args.batch_size, 1, 584, 584])

    print("Starting training")
    for i in range(args.epochs):
        print("\nEpoch %d/%d" % (i + 1, args.epochs))
        eddl.reset_loss(segnet)
        for j in range(args.num_batches):
            eddl.next_batch([x_train, y_train], [xbatch, ybatch])
            # DA net
            eddl.forward(danet, [xbatch, ybatch])
            xbatch_da = eddl.getOutput(img)
            ybatch_da = eddl.getOutput(mask)
            # SegNet
            eddl.train_batch(segnet, [xbatch_da], [ybatch_da])
            eddl.print_loss(segnet, j)
            if i == args.epochs - 1:
                yout = eddl.getOutput(out).select(["0"])
                yout.save("./out_%d.jpg" % j)
            print()
    print("All done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--epochs", type=int, metavar="INT", default=10)
    parser.add_argument("--batch-size", type=int, metavar="INT", default=1)
    parser.add_argument("--num-batches", type=int, metavar="INT", default=5)
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--mem", metavar="|".join(MEM_CHOICES),
                        choices=MEM_CHOICES, default="low_mem")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import pyeddl._core.eddl as eddl_core
import pyeddl.eddl as eddl_py


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_core_layers(eddl):
    in2d = eddl.Input([16])
    in3d = eddl.Input([1, 16])
    in4d = eddl.Input([3, 16, 16])
    eddl.Activation(in2d, "relu")
    eddl.Activation(in2d, "relu", [0.02])
    eddl.Activation(in2d, "relu", [0.02], "foo")
    eddl.Softmax(in2d)
    eddl.Softmax(in2d, 0)
    eddl.Softmax(in2d, 0, "foo")
    eddl.Sigmoid(in2d)
    eddl.Sigmoid(in2d, "foo")
    eddl.HardSigmoid(in2d)
    eddl.HardSigmoid(in2d, "foo")
    eddl.ReLu(in2d)
    eddl.ReLu(in2d, "foo")
    eddl.ThresholdedReLu(in2d)
    eddl.ThresholdedReLu(in2d, 1.0)
    eddl.ThresholdedReLu(in2d, 1.0, "foo")
    eddl.LeakyReLu(in2d)
    eddl.LeakyReLu(in2d, 0.02)
    eddl.LeakyReLu(in2d, 0.02, "foo")
    eddl.Elu(in2d)
    eddl.Elu(in2d, 1.0)
    eddl.Elu(in2d, 1.0, "foo")
    eddl.Selu(in2d)
    eddl.Selu(in2d, "foo")
    eddl.Exponential(in2d)
    eddl.Exponential(in2d, "foo")
    eddl.Softplus(in2d)
    eddl.Softplus(in2d, "foo")
    eddl.Softsign(in2d)
    eddl.Softsign(in2d, "foo")
    eddl.Linear(in2d)
    eddl.Linear(in2d, 1.0)
    eddl.Linear(in2d, 1.0, "foo")
    eddl.Tanh(in2d)
    eddl.Tanh(in2d, "foo")
    eddl.Conv(in4d, 16, [1, 1])
    eddl.Conv(in4d, 16, [1, 1], [2, 2], "none")
    eddl.Conv(in4d, 16, [1, 1], [2, 2], "none", True)
    eddl.Conv(in4d, 16, [1, 1], [2, 2], "none", True, 1)
    eddl.Conv(in4d, 16, [1, 1], [2, 2], "none", True, 1, [1, 1])
    eddl.Conv(in4d, 16, [1, 1], [2, 2], "none", True, 1, [1, 1], "foo")
    eddl.Conv1D(in3d, 16, [3])
    eddl.Conv1D(in3d, 16, [3], [1])
    eddl.Conv1D(in3d, 16, [3], [1], "none")
    eddl.Conv1D(in3d, 16, [3], [1], "none", False)
    eddl.Conv1D(in3d, 16, [3], [1], "none", False, 1)
    eddl.Conv1D(in3d, 16, [3], [1], "none", False, 1, [1])
    eddl.Conv1D(in3d, 16, [3], [1], "none", False, 1, [1], "foo")
    eddl.PointwiseConv(in4d, 16, [1, 1])
    eddl.PointwiseConv(in4d, 16, [1, 1], False)
    eddl.PointwiseConv(in4d, 16, [1, 1], False, 1)
    eddl.PointwiseConv(in4d, 16, [1, 1], False, 1, [1, 1])
    eddl.PointwiseConv(in4d, 16, [1, 1], False, 1, [1, 1], "foo")
    eddl.ConvT(in4d, 16, [1, 1], [2, 2])
    eddl.ConvT(in4d, 16, [1, 1], [2, 2], "none")
    eddl.ConvT(in4d, 16, [1, 1], [2, 2], "none", [1, 1])
    eddl.ConvT(in4d, 16, [1, 1], [2, 2], "none", [1, 1], [1, 1])
    eddl.ConvT(in4d, 16, [1, 1], [2, 2], "none", [1, 1], [1, 1], True)
    eddl.ConvT(in4d, 16, [1, 1], [2, 2], "none", [1, 1], [1, 1], True, "foo")
    eddl.Dense(in2d, 16)
    eddl.Dense(in2d, 16, True)
    eddl.Dense(in2d, 16, True, "foo")
    eddl.Embedding(in2d, 1001, 16, 250)
    eddl.Embedding(in2d, 1001, 16, 250, True)
    eddl.Embedding(in2d, 1001, 16, 250, True, "foo")
    eddl.Input([16], "foo")
    eddl.UpSampling(in4d, [2, 2])
    eddl.UpSampling(in4d, [2, 2], "nearest")
    eddl.UpSampling(in4d, [2, 2], "nearest", "foo")
    eddl.Reshape(in2d, [1, 4, 4])
    eddl.Reshape(in2d, [1, 4, 4], "foo")
    eddl.Flatten(in2d)
    eddl.Flatten(in2d, "foo")
    eddl.Transpose(in2d)
    eddl.Transpose(in2d, "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_transformations(eddl):
    in2d = eddl.Input([16])
    eddl.Shift(in2d, [1, 1])
    eddl.Shift(in2d, [1, 1], "nearest")
    eddl.Shift(in2d, [1, 1], "nearest", 0.0)
    eddl.Shift(in2d, [1, 1], "nearest", 0.0, "foo")
    eddl.Rotate(in2d, 1.1)
    eddl.Rotate(in2d, 1.1, [0, 0])
    eddl.Rotate(in2d, 1.1, [0, 0], "nearest")
    eddl.Rotate(in2d, 1.1, [0, 0], "nearest", 0.0)
    eddl.Rotate(in2d, 1.1, [0, 0], "nearest", 0.0, "foo")
    eddl.Scale(in2d, [1, 4], True)
    eddl.Scale(in2d, [1, 4], True, "nearest")
    eddl.Scale(in2d, [1, 4], True, "nearest", 0.0)
    eddl.Scale(in2d, [1, 4], True, "nearest", 0.0, "foo")
    eddl.Flip(in2d)
    eddl.Flip(in2d, 0)
    eddl.Flip(in2d, 0, "foo")
    eddl.HorizontalFlip(in2d)
    eddl.HorizontalFlip(in2d, "foo")
    eddl.VerticalFlip(in2d)
    eddl.VerticalFlip(in2d, "foo")
    eddl.Crop(in2d, [0, 0], [3, 3])
    eddl.Crop(in2d, [0, 0], [3, 3], True)
    eddl.Crop(in2d, [0, 0], [3, 3], True, 0.0)
    eddl.Crop(in2d, [0, 0], [3, 3], True, 0.0, "foo")
    eddl.CenteredCrop(in2d, [3, 3])
    eddl.CenteredCrop(in2d, [3, 3], True)
    eddl.CenteredCrop(in2d, [3, 3], True, 0.0)
    eddl.CenteredCrop(in2d, [3, 3], True, 0.0, "foo")
    eddl.CropScale(in2d, [0, 0], [3, 3])
    eddl.CropScale(in2d, [0, 0], [3, 3], "nearest")
    eddl.CropScale(in2d, [0, 0], [3, 3], "nearest", 0.0)
    eddl.CropScale(in2d, [0, 0], [3, 3], "nearest", 0.0, "foo")
    eddl.Cutout(in2d, [0, 0], [3, 3])
    eddl.Cutout(in2d, [0, 0], [3, 3], 0.0)
    eddl.Cutout(in2d, [0, 0], [3, 3], 0.0, "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_data_augmentation(eddl):
    in2d = eddl.Input([16])
    eddl.RandomShift(in2d, [1.0, 1.0], [2.0, 2.0])
    eddl.RandomShift(in2d, [1.0, 1.0], [2.0, 2.0], "nearest")
    eddl.RandomShift(in2d, [1.0, 1.0], [2.0, 2.0], "nearest", 0.0)
    eddl.RandomShift(in2d, [1.0, 1.0], [2.0, 2.0], "nearest", 0.0, "foo")
    eddl.RandomRotation(in2d, [1.0, 1.0])
    eddl.RandomRotation(in2d, [1.0, 1.0], [0, 0])
    eddl.RandomRotation(in2d, [1.0, 1.0], [0, 0], "nearest")
    eddl.RandomRotation(in2d, [1.0, 1.0], [0, 0], "nearest", 0.0)
    eddl.RandomRotation(in2d, [1.0, 1.0], [0, 0], "nearest", 0.0, "foo")
    eddl.RandomScale(in2d, [1.0, 1.0])
    eddl.RandomScale(in2d, [1.0, 1.0], "nearest")
    eddl.RandomScale(in2d, [1.0, 1.0], "nearest", 0.0)
    eddl.RandomScale(in2d, [1.0, 1.0], "nearest", 0.0, "foo")
    eddl.RandomFlip(in2d, 0)
    eddl.RandomFlip(in2d, 0, "foo")
    eddl.RandomHorizontalFlip(in2d)
    eddl.RandomHorizontalFlip(in2d, "foo")
    eddl.RandomVerticalFlip(in2d)
    eddl.RandomVerticalFlip(in2d, "foo")
    eddl.RandomCrop(in2d, [4, 4])
    eddl.RandomCrop(in2d, [4, 4], "foo")
    eddl.RandomCropScale(in2d, [1.0, 1.0])
    eddl.RandomCropScale(in2d, [1.0, 1.0], "nearest")
    eddl.RandomCropScale(in2d, [1.0, 1.0], "nearest", "foo")
    eddl.RandomCutout(in2d, [1.0, 1.0], [1.0, 1.0])
    eddl.RandomCutout(in2d, [1.0, 1.0], [1.0, 1.0], 0.0)
    eddl.RandomCutout(in2d, [1.0, 1.0], [1.0, 1.0], 0.0, "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_losses(eddl):
    eddl.getLoss("mse")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_metrics(eddl):
    eddl.getMetric("mse")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_merge_layers(eddl):
    in_1 = eddl.Input([16])
    in_2 = eddl.Input([16])
    eddl.Add([in_1, in_2])
    eddl.Add([in_1, in_2], "foo")
    eddl.Average([in_1, in_2])
    eddl.Average([in_1, in_2], "foo")
    eddl.Concat([in_1, in_2])
    eddl.Concat([in_1, in_2], 1)
    eddl.Concat([in_1, in_2], 1, "foo")
    eddl.MatMul([in_1, in_2])
    eddl.MatMul([in_1, in_2], "foo")
    eddl.Maximum([in_1, in_2])
    eddl.Maximum([in_1, in_2], "foo")
    eddl.Minimum([in_1, in_2])
    eddl.Minimum([in_1, in_2], "foo")
    eddl.Subtract([in_1, in_2])
    eddl.Subtract([in_1, in_2], "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_noise_layers(eddl):
    in2d = eddl.Input([16])
    eddl.GaussianNoise(in2d, 1.1)
    eddl.GaussianNoise(in2d, 1.1, "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_normalization_layers(eddl):
    in2d = eddl.Input([16])
    eddl.BatchNormalization(in2d, True)
    eddl.BatchNormalization(in2d, True, 0.9)
    eddl.BatchNormalization(in2d, True, 0.9, 0.001)
    eddl.BatchNormalization(in2d, True, 0.9, 0.001)
    eddl.BatchNormalization(in2d, True, 0.9, 0.001, "foo")
    eddl.LayerNormalization(in2d, True)
    eddl.LayerNormalization(in2d, True, 0.001)
    eddl.LayerNormalization(in2d, True, 0.001, "foo")
    # Norm{,Max,MinMax} tests now lead to a "RuntimeError: axis 1 >= dim=1"
    # eddl.Norm(in2d)
    # eddl.Norm(in2d, 0.001)
    # eddl.Norm(in2d, 0.001, "foo")
    # eddl.NormMax(in2d)
    # eddl.NormMax(in2d, 0.001)
    # eddl.NormMax(in2d, 0.001, "foo")
    # eddl.NormMinMax(in2d)
    # eddl.NormMinMax(in2d, 0.001)
    # eddl.NormMinMax(in2d, 0.001, "foo")
    eddl.Dropout(in2d, 0.5)
    eddl.Dropout(in2d, 0.5, False, "foo")
    eddl.Dropout(in2d, 0.5, False, "foo")
    in4d = eddl.Input([6, 16, 16])
    eddl.GroupNormalization(in4d, 3)
    eddl.GroupNormalization(in4d, 3, 0.001)
    eddl.GroupNormalization(in4d, 3, 0.001, True)
    eddl.GroupNormalization(in4d, 3, 0.001, True, "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_operator_layers(eddl):
    in_1 = eddl.Input([16])
    in_2 = eddl.Input([16])
    eddl.Abs(in_1)
    # Diff is deprecated in favor of Sub
    # eddl.Diff(in_1, in_2)
    # eddl.Diff(in_1, 1.0)
    # eddl.Diff(1.0, in_1)
    eddl.Sub(in_1, in_2)
    eddl.Sub(in_1, 1.0)
    eddl.Sub(1.0, in_1)
    eddl.Div(in_1, in_2)
    eddl.Div(in_1, 1.0)
    eddl.Div(1.0, in_1)
    eddl.Exp(in_1)
    eddl.Log(in_1)
    eddl.Log2(in_1)
    eddl.Log10(in_1)
    eddl.Mult(in_1, in_2)
    eddl.Mult(in_1, 1.0)
    eddl.Mult(1.0, in_1)
    eddl.Pow(in_1, in_2)
    eddl.Pow(in_1, 1.0)
    eddl.Sqrt(in_1)
    # Sum is deprecated in favor of Add
    # eddl.Sum(in_1, in_2)
    # eddl.Sum(in_1, 1.0)
    # eddl.Sum(1.0, in_1)
    eddl.Add(in_1, in_2)
    eddl.Add(in_1, 1.0)
    eddl.Add(1.0, in_1)

    in4d = eddl.Input([3, 16, 16])
    eddl.Select(in4d, [":", ":8", ":8"])
    eddl.Permute(in4d, [0, 1, 2])


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_reduction_layers(eddl):
    in2d = eddl.Input([16])
    eddl.ReduceMean(in2d, [0])
    eddl.ReduceMean(in2d, [0], False)
    eddl.ReduceVar(in2d, [0])
    eddl.ReduceVar(in2d, [0], False)
    eddl.ReduceSum(in2d, [0])
    eddl.ReduceSum(in2d, [0], False)
    eddl.ReduceMax(in2d, [0])
    eddl.ReduceMax(in2d, [0], False)
    eddl.ReduceMin(in2d, [0])
    eddl.ReduceMin(in2d, [0], False)
    eddl.ReduceArgMax(in2d, [0])
    eddl.ReduceArgMax(in2d, [0], False)


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_generator_layers(eddl):
    eddl.GaussGenerator(0.0, 1.0, [10])
    eddl.UniformGenerator(0.0, 5.0, [10])


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_optimizers(eddl):
    eddl.adadelta(0.01, 0.9, 0.0001, 0.0)
    eddl.adam()
    eddl.adam(0.01)
    eddl.adam(0.01, 0.9)
    eddl.adam(0.01, 0.9, 0.999)
    eddl.adam(0.01, 0.9, 0.999, 0.0001)
    eddl.adam(0.01, 0.9, 0.999, 0.0001, 0.0)
    eddl.adam(0.01, 0.9, 0.999, 0.0001, 0.0, False)
    eddl.adagrad(0.01, 0.0001, 0.0)
    eddl.adamax(0.01, 0.9, 0.999, 0.0001, 0.0)
    eddl.nadam(0.01, 0.9, 0.999, 0.0001, 0.0)
    eddl.rmsprop()
    eddl.rmsprop(0.01)
    eddl.rmsprop(0.01, 0.9)
    eddl.rmsprop(0.01, 0.9, 0.0001)
    eddl.rmsprop(0.01, 0.9, 0.0001, 0.0)
    eddl.sgd()
    eddl.sgd(0.01)
    eddl.sgd(0.01, 0.0)
    eddl.sgd(0.01, 0.0, 0.0)
    eddl.sgd(0.01, 0.0, 0.0, False)


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_pooling_layers(eddl):
    in4d = eddl.Input([3, 16, 16])
    eddl.AveragePool(in4d)
    eddl.AveragePool(in4d, [2, 2])
    eddl.AveragePool(in4d, [2, 2], [2, 2])
    eddl.AveragePool(in4d, [2, 2], [2, 2], "none")
    eddl.AveragePool(in4d, [2, 2], [2, 2], "none", "foo")
    eddl.MaxPool(in4d)
    eddl.MaxPool(in4d, [2, 2])
    eddl.MaxPool(in4d, [2, 2], [2, 2])
    eddl.MaxPool(in4d, [2, 2], [2, 2], "none")
    eddl.MaxPool(in4d, [2, 2], [2, 2], "none", "foo")
    eddl.GlobalMaxPool(in4d)
    eddl.GlobalMaxPool(in4d, "foo")
    eddl.GlobalAveragePool(in4d)
    eddl.GlobalAveragePool(in4d, "foo")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_recurrent_layers(eddl):
    in2d = eddl.Input([16])
    eddl.RNN(in2d, 1)
    eddl.RNN(in2d, 1, "tanh")
    eddl.RNN(in2d, 1, "tanh", True)
    eddl.RNN(in2d, 1, "tanh", True, True)
    eddl.RNN(in2d, 1, "tanh", True, True, "foo")
    eddl.LSTM(in2d, 1)
    eddl.LSTM(in2d, 1, True)
    eddl.LSTM(in2d, 1, True, True)
    eddl.LSTM(in2d, 1, True, True, "foo")
    # Decoder
    enc = eddl.LSTM(in2d, 1, True)
    eddl.Decoder(eddl.LSTM(in2d, 1), enc)
    eddl.Decoder(eddl.LSTM(in2d, 1), enc, "concat")


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_initializers(eddl):
    in2d = eddl.Input([16])
    eddl.GlorotNormal(in2d)
    eddl.GlorotNormal(in2d, 1234)
    eddl.GlorotUniform(in2d)
    eddl.GlorotUniform(in2d, 1234)
    eddl.HeNormal(in2d)
    eddl.HeNormal(in2d, 1234)
    eddl.HeUniform(in2d)
    eddl.HeUniform(in2d, 1234)
    eddl.RandomNormal(in2d)
    eddl.RandomNormal(in2d, 0.0)
    eddl.RandomNormal(in2d, 0.0, 0.1)
    eddl.RandomNormal(in2d, 0.0, 0.1, 1234)
    eddl.RandomUniform(in2d)
    eddl.RandomUniform(in2d, 0.0)
    eddl.RandomUniform(in2d, 0.0, 0.1)
    eddl.RandomUniform(in2d, 0.0, 0.1, 1234)
    eddl.Constant(in2d)
    eddl.Constant(in2d, 0.1)


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_regularizers(eddl):
    in2d = eddl.Input([16])
    eddl.L2(in2d, 0.001)
    eddl.L1(in2d, 0.001)
    eddl.L1L2(in2d, 0.001, 0.001)


@pytest.mark.parametrize("eddl", [eddl_core, eddl_py])
def test_computing_services(eddl):
    eddl.CS_CPU()
    eddl.CS_CPU(1)
    eddl.CS_CPU(1, "low_mem")
    if eddl is eddl_py:
        eddl.CS_GPU()
    eddl.CS_GPU([1])
    eddl.CS_GPU([1], 1)
    eddl.CS_GPU([1], 1, "low_mem")
    eddl.CS_FPGA([1])
    eddl.CS_FPGA([1], 1)
    eddl.CS_COMPSS("foo.xml")



# Copyright (c) 2019-2020 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import numpy as np
from pyeddl._core import Tensor as CoreTensor
from pyeddl.tensor import Tensor as PyTensor
from pyeddl._core.eddl import DEV_CPU


# --- Creation ---

@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_create(Tensor):
    shape = [2, 3]
    t = Tensor(shape)
    assert t.shape == shape
    assert t.isCPU()
    t = Tensor(shape, DEV_CPU)
    assert t.shape == shape
    assert t.isCPU()


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_array_from_tensor(Tensor):
    shape = [2, 3]
    t = Tensor(shape)
    t.fill_(1.0)
    a = np.array(t, copy=False)
    b = np.ones(shape, dtype=np.float32)
    assert np.array_equal(a, b)
    c = np.array(t, copy=True)
    assert np.array_equal(c, b)
    d = t.getdata()
    assert np.array_equal(d, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_tensor_from_array(Tensor):
    R, C = 2, 3
    a = np.arange(6).reshape([R, C]).astype(np.float32)
    t = Tensor(a)
    assert t.shape == [R, C]
    b = np.array(t, copy=True)
    assert np.array_equal(a, b)
    # check creation from 1D array
    a = np.array([1, 2]).astype(np.float32)
    t = Tensor(a)
    b = np.array(t, copy=True)
    assert np.array_equal(a, b)
    # check automatic type conversion
    a = np.array([[1, 2], [3, 4]])
    t = Tensor(a)
    b = np.array(t, copy=True).astype(a.dtype)
    assert np.array_equal(a, b)
    # check fromarray
    if Tensor is PyTensor:
        a = np.array([1, 2]).astype(np.float32)
        t = Tensor.fromarray(a)
        b = np.array(t, copy=True)
        assert np.array_equal(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_tensor_array_ops(Tensor):
    a = np.arange(6).reshape([2, 3]).astype(np.float32)
    incr = 2.0
    b = a + incr
    t = Tensor(list(a.shape))
    t.fill_(incr)
    a += t
    assert np.array_equal(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_ones(Tensor):
    shape = [2, 3]
    t = Tensor.ones(shape)
    a = np.array(t, copy=False)
    b = np.ones(shape, dtype=np.float32)
    assert np.array_equal(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_zeros(Tensor):
    shape = [2, 3]
    t = Tensor.zeros(shape)
    a = np.array(t, copy=False)
    b = np.zeros(shape, dtype=np.float32)
    assert np.array_equal(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_full(Tensor):
    shape = [2, 3]
    value = 42
    t = Tensor.full(shape, value)
    a = np.array(t, copy=False)
    b = np.full(shape, value, dtype=np.float32)
    assert np.array_equal(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_arange(Tensor):
    start, stop, step = 0, 2, .33
    t = Tensor.arange(start, stop, step)
    a = np.array(t, copy=False)
    b = np.arange(start, stop, step, dtype=np.float32)
    assert np.allclose(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_range(Tensor):
    start, stop, step = 0, 2, .33
    t = Tensor.range(start, stop, step)
    a = np.array(t, copy=False)
    b = np.arange(start, stop, step, dtype=np.float32)
    assert np.allclose(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_linspace(Tensor):
    start, stop, num = 0, 2, 7
    t = Tensor.linspace(start, stop, num)
    a = np.array(t, copy=False)
    b = np.linspace(start, stop, num, dtype=np.float32)
    assert np.allclose(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_logspace(Tensor):
    start, stop, num, base = 0.1, 1.0, 5, 10.0
    t = Tensor.logspace(start, stop, num, base)
    a = np.array(t, copy=False)
    b = np.logspace(start, stop, num, dtype=np.float32)
    assert np.allclose(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_eye(Tensor):
    size = 3
    t = Tensor.eye(size)
    a = np.array(t, copy=False)
    assert np.array_equal(a.diagonal(), np.ones(size, dtype=np.float32))
    u = Tensor.eye(size, 1)
    b = np.array(u, copy=False)
    assert(np.array_equal(b, np.eye(size, k=1)))


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_randn(Tensor):
    shape = [2, 3]
    t = Tensor.randn(shape)
    assert Tensor.getShape(t) == shape


# --- Copy ---

# toGPU fails if EDDL is compiled for GPU but no CUDA devices are
# detected (CUDA error 100) or there are driver issues (CUDA error 35).
# To conditionally skip this we probably  need some support from EDDL

# @pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
# def test_cpu_gpu(Tensor):
#     # toGPU and toCPU are no-ops if EDDL is not compiled for GPU
#     a = np.arange(6).reshape(2, 3).astype(np.float32)
#     t = Tensor(a)
#     t.toGPU()
#     t.toCPU()
#     b = np.array(t, copy=False)
#     assert np.array_equal(b, a)
#     # check getdata from GPU tensor
#     t = Tensor([2, 3], DEV_CPU)
#     t.fill_(2)
#     t.toGPU()
#     a = t.getdata()
#     assert np.alltrue(a == 2)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clone(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    t = Tensor(a)
    u = t.clone()
    b = np.array(u, copy=False)
    assert np.array_equal(b, a)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_select(Tensor):
    R, C = 3, 4
    a = np.arange(R * C).reshape(R, C).astype(np.float32)
    t = Tensor(a)
    u = t.select(["1:3", "2"])
    b = np.array(u, copy=False)
    # numpy selection converts to row vector in this case
    assert np.array_equal(b, a[1:3, 2][:, np.newaxis])


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_copy(Tensor):
    R, C = 3, 4
    a = np.arange(R * C).reshape(R, C).astype(np.float32)
    t = Tensor(a)
    u = Tensor([R, C])
    Tensor.copy(t, u)
    b = np.array(u, copy=False)
    assert np.array_equal(b, a)


# --- Serialization ---

@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_load_save(Tensor, tmp_path):
    fn = str(tmp_path / "tensor.bin")
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    t = Tensor(a)
    t.save(fn, "bin")
    u = Tensor.load(fn, "bin")
    b = np.array(u, copy=False)
    assert np.array_equal(b, a)


# --- Math ---

@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_abs_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.abs_()
    b = np.array(t, copy=False)
    assert np.allclose(np.abs(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_abs(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.abs(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.abs(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_acos_(Tensor):
    t = Tensor.range(-1, 1, .5)
    a = np.array(t, copy=True)
    t.acos_()
    b = np.array(t, copy=False)
    assert np.allclose(np.arccos(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_acos(Tensor):
    t = Tensor.range(-1, 1, .5)
    a = np.array(t, copy=False)
    u = Tensor.acos(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.arccos(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_add_(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    n = 2
    # add scalar to tensor
    t = Tensor(a)
    t.add_(n)
    b = np.array(t, copy=False)
    assert np.allclose(a + n, b)
    # add tensor to tensor
    t = Tensor(a)
    c = 2 * a
    u = Tensor(c)
    t.add_(u)
    b = np.array(t, copy=False)
    assert np.allclose(b, a + c)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_add(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    n = 2
    # tensor = tensor + scalar
    t = Tensor(a)
    u = Tensor.add(t, n)
    b = np.array(u, copy=False)
    assert np.allclose(a + n, b)
    # tensor = tensor + tensor
    t = Tensor(a)
    u = Tensor(a + n)
    v = Tensor.add(t, u)
    b = np.array(v, copy=False)
    assert np.allclose(b, a + a + n)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_asin_(Tensor):
    t = Tensor.range(-1, 1, .5)
    a = np.array(t, copy=True)
    t.asin_()
    b = np.array(t, copy=False)
    assert np.allclose(np.arcsin(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_asin(Tensor):
    t = Tensor.range(-1, 1, .5)
    a = np.array(t, copy=False)
    u = Tensor.asin(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.arcsin(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_atan_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.atan_()
    b = np.array(t, copy=False)
    assert np.allclose(np.arctan(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_atan(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.atan(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.arctan(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_ceil_(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=True)
    t.ceil_()
    b = np.array(t, copy=False)
    assert np.allclose(np.ceil(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_ceil(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=False)
    u = Tensor.ceil(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.ceil(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clamp_(Tensor):
    low, high = -1, 1
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=True)
    t.clamp_(low, high)
    b = np.array(t, copy=False)
    assert np.allclose(np.clip(a, low, high), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clamp(Tensor):
    low, high = -1, 1
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=False)
    u = Tensor.clamp(t, low, high)
    b = np.array(u, copy=False)
    assert np.allclose(np.clip(a, low, high), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clampmax_(Tensor):
    high = 1
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=True)
    t.clampmax_(high)
    b = np.array(t, copy=False)
    assert np.allclose(np.clip(a, None, high), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clampmax(Tensor):
    high = 1
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=False)
    u = Tensor.clampmax(t, high)
    b = np.array(u, copy=False)
    assert np.allclose(np.clip(a, None, high), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clampmin_(Tensor):
    low = -1
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=True)
    t.clampmin_(low)
    b = np.array(t, copy=False)
    assert np.allclose(np.clip(a, low, None), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_clampmin(Tensor):
    low = -1
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=False)
    u = Tensor.clampmin(t, low)
    b = np.array(u, copy=False)
    assert np.allclose(np.clip(a, low, None), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_cos_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.cos_()
    b = np.array(t, copy=False)
    assert np.allclose(np.cos(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_cos(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.cos(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.cos(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_cosh_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.cosh_()
    b = np.array(t, copy=False)
    assert np.allclose(np.cosh(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_cosh(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.cosh(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.cosh(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_diag_(Tensor):
    a = np.arange(16).reshape(4, 4).astype(np.float32)
    for k in range(-3, 4):
        t = Tensor(a)
        t.diag_(k=k)
        b = np.array(t, copy=False)
        exp = np.diagflat(np.diag(a, k=k), k=k)
        assert np.allclose(exp, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_diag(Tensor):
    a = np.arange(16).reshape(4, 4).astype(np.float32)
    t = Tensor(a)
    for k in range(-3, 4):
        u = t.diag(k=k)
        b = np.array(u, copy=False)
        exp = np.diagflat(np.diag(a, k=k), k=k)
        assert np.allclose(exp, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_div_(Tensor):
    a = np.arange(2, 8).reshape(2, 3).astype(np.float32)
    n = 2
    # divide scalar by tensor
    t = Tensor(a)
    t.div_(n)
    b = np.array(t, copy=False)
    assert np.allclose(a / n, b)
    # divide tensor by tensor
    t = Tensor(a)
    c = 2 * a
    u = Tensor(c)
    t.div_(u)
    b = np.array(t, copy=False)
    assert np.allclose(b, a / c)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_div(Tensor):
    a = np.arange(2, 8).reshape(2, 3).astype(np.float32)
    n = 2
    # tensor = tensor / scalar
    t = Tensor(a)
    u = Tensor.div(t, n)
    b = np.array(u, copy=False)
    assert np.allclose(a / n, b)
    # tensor = tensor / tensor
    t = Tensor(a)
    u = Tensor(a + n)
    v = Tensor.div(t, u)
    b = np.array(v, copy=False)
    assert np.allclose(b, a / (a + n))


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_exp_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.exp_()
    b = np.array(t, copy=False)
    assert np.allclose(np.exp(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_exp(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.exp(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.exp(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_floor_(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=True)
    t.floor_()
    b = np.array(t, copy=False)
    assert np.allclose(np.floor(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_floor(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.25)
    a = np.array(t, copy=False)
    u = Tensor.floor(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.floor(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_log_(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.log_()
    b = np.array(t, copy=False)
    assert np.allclose(np.log(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_log(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.log(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.log(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_log2_(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.log2_()
    b = np.array(t, copy=False)
    assert np.allclose(np.log2(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_log2(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.log2(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.log2(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_log10_(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.log10_()
    b = np.array(t, copy=False)
    assert np.allclose(np.log10(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_log10(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.log10(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.log10(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_logn_(Tensor):
    base = 3
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.logn_(base)
    b = np.array(t, copy=False)
    assert np.allclose(np.log(a) / np.log(base), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_logn(Tensor):
    base = 3
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.logn(t, base)
    b = np.array(u, copy=False)
    assert np.allclose(np.log(a) / np.log(base), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_max_min(Tensor):
    start, stop = -4, 3
    t = Tensor.range(start, stop, 1)
    assert t.max() == stop
    assert t.min() == start


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_mod_(Tensor):
    n = 2
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.mod_(n)
    b = np.array(t, copy=False)
    assert np.allclose(np.mod(a, n), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_mod(Tensor):
    n = 2
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.mod(t, n)
    b = np.array(u, copy=False)
    assert np.allclose(np.mod(a, n), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_mult_(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    n = 2
    # multiply scalar by tensor
    t = Tensor(a)
    t.mult_(n)
    b = np.array(t, copy=False)
    assert np.allclose(a * n, b)
    # multiply tensor by tensor
    t = Tensor(a)
    c = 2 + a
    u = Tensor(c)
    t.mult_(u)
    b = np.array(t, copy=False)
    assert np.allclose(b, a * c)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_mult(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    n = 2
    # tensor = tensor * scalar
    t = Tensor(a)
    u = Tensor.mult(t, n)
    b = np.array(u, copy=False)
    assert np.allclose(a * n, b)
    # tensor = tensor * tensor
    t = Tensor(a)
    u = Tensor(a + n)
    v = Tensor.mult(t, u)
    b = np.array(v, copy=False)
    assert np.allclose(b, a * (a + n))


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_mult2D(Tensor):
    a = np.arange(2, 8).reshape(2, 3).astype(np.float32)
    b = np.arange(1, 13).reshape(3, 4).astype(np.float32)
    t = Tensor(a)
    u = Tensor(b)
    v = Tensor.mult2D(t, u)
    c = np.array(v, copy=False)
    assert np.allclose(c, a @ b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_normalize_(Tensor):
    m, M = -1, 1
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.normalize_(m, M)
    b = np.array(t, copy=False)
    r = (M - m) / (a.max() - a.min())
    assert np.allclose(r * (a - a.min()) + m, b, atol=1e-7)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_normalize(Tensor):
    m, M = -1, 1
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.normalize(t, m, M)
    b = np.array(u, copy=False)
    r = (M - m) / (a.max() - a.min())
    assert np.allclose(r * (a - a.min()) + m, b, atol=1e-7)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_neg_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.neg_()
    b = np.array(t, copy=False)
    assert np.allclose(-a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_neg(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.neg(t)
    b = np.array(u, copy=False)
    assert np.allclose(-a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_reciprocal_(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.reciprocal_()
    b = np.array(t, copy=False)
    assert np.allclose(np.reciprocal(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_reciprocal(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.reciprocal(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.reciprocal(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_round_(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.4)
    a = np.array(t, copy=True)
    t.round_()
    b = np.array(t, copy=False)
    assert np.allclose(np.round(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_round(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.4)
    a = np.array(t, copy=False)
    u = Tensor.round(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.round(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_rsqrt_(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.rsqrt_()
    b = np.array(t, copy=False)
    assert np.allclose(1 / np.sqrt(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_rsqrt(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.rsqrt(t)
    b = np.array(u, copy=False)
    assert np.allclose(1 / np.sqrt(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sigmoid_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.sigmoid_()
    b = np.array(t, copy=False)
    assert np.allclose(1 / (1 + np.exp(-a)), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sigmoid(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.sigmoid(t)
    b = np.array(u, copy=False)
    assert np.allclose(1 / (1 + np.exp(-a)), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sign_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.sign_()
    b = np.array(t, copy=False)
    assert np.allclose(np.sign(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sign(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.sign(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.sign(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sin_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.sin_()
    b = np.array(t, copy=False)
    assert np.allclose(np.sin(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sin(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.sin(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.sin(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sinh_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.sinh_()
    b = np.array(t, copy=False)
    assert np.allclose(np.sinh(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sinh(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.sinh(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.sinh(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sqr_(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=True)
    t.sqr_()
    b = np.array(t, copy=False)
    assert np.allclose(np.square(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sqr(Tensor):
    t = Tensor.range(-5, 5, 1)
    a = np.array(t, copy=False)
    u = Tensor.sqr(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.square(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sqrt_(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=True)
    t.sqrt_()
    b = np.array(t, copy=False)
    assert np.allclose(np.sqrt(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sqrt(Tensor):
    t = Tensor.range(1, 10, 1)
    a = np.array(t, copy=False)
    u = Tensor.sqrt(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.sqrt(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sub_(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    n = 2
    # subtract scalar from tensor
    t = Tensor(a)
    t.sub_(n)
    b = np.array(t, copy=False)
    assert np.allclose(a - n, b)
    # subtract tensor from tensor
    t = Tensor(a)
    c = 2 * a
    u = Tensor(c)
    t.sub_(u)
    b = np.array(t, copy=False)
    assert np.allclose(b, a - c)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_sub(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    n = 2
    # tensor = tensor - scalar
    t = Tensor(a)
    u = Tensor.sub(t, n)
    b = np.array(u, copy=False)
    assert np.allclose(a - n, b)
    # tensor = tensor - tensor
    t = Tensor(a)
    u = Tensor(a / n)
    v = Tensor.sub(t, u)
    b = np.array(v, copy=False)
    assert np.allclose(b, a - (a / n))


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_tan_(Tensor):
    t = Tensor.range(-1, 1, .2)
    a = np.array(t, copy=True)
    t.tan_()
    b = np.array(t, copy=False)
    assert np.allclose(np.tan(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_tan(Tensor):
    t = Tensor.range(-1, 1, .2)
    a = np.array(t, copy=False)
    u = Tensor.tan(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.tan(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_tanh_(Tensor):
    t = Tensor.range(-1, 1, .2)
    a = np.array(t, copy=True)
    t.tanh_()
    b = np.array(t, copy=False)
    assert np.allclose(np.tanh(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_tanh(Tensor):
    t = Tensor.range(-1, 1, .2)
    a = np.array(t, copy=False)
    u = Tensor.tanh(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.tanh(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_trace(Tensor):
    a = np.arange(16).reshape(4, 4).astype(np.float32)
    t = Tensor(a)
    for k in range(-3, 4):
        assert np.allclose(t.trace(k=k), np.trace(a, offset=k))


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_trunc_(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.4)
    a = np.array(t, copy=True)
    t.trunc_()
    b = np.array(t, copy=False)
    assert np.allclose(np.trunc(a), b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_trunc(Tensor):
    t = Tensor.range(-2.0, 2.0, 0.4)
    a = np.array(t, copy=False)
    u = Tensor.trunc(t)
    b = np.array(u, copy=False)
    assert np.allclose(np.trunc(a), b)


# --- Transformations ---

@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_scale(Tensor):
    shape = [10, 3, 128, 128]
    t = Tensor(shape)
    target_shape = [64, 64]
    u = t.scale(target_shape)
    assert u.shape == shape[:2] + target_shape
    u = t.scale(target_shape, keep_size=True)
    assert u.shape == shape


# --- Other ---

@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_fill_(Tensor):
    shape = [2, 3]
    value = 42
    t = Tensor(shape)
    t.fill_(value)
    a = np.array(t, copy=False)
    b = np.full(shape, value, dtype=np.float32)
    assert np.array_equal(a, b)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_reshape_(Tensor):
    R, C = 3, 4
    a = np.arange(R * C).reshape(R, C).astype(np.float32)
    t = Tensor(a)
    t.reshape_([C, R])
    b = np.array(t, copy=False)
    c = a.reshape(C, R)
    assert np.array_equal(b, c)


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_info_print(Tensor):
    a = np.arange(6).reshape(2, 3).astype(np.float32)
    t = Tensor(a)
    t.info()
    t.print()


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_getShape(Tensor):
    shape = [3, 4]
    t = Tensor(shape)
    assert t.getShape() == shape


@pytest.mark.parametrize("Tensor", [CoreTensor, PyTensor])
def test_onehot(Tensor):
    a = np.array([
        [0, 2, 4],
        [2, 3, 0]
    ])
    vocs = 5
    t = Tensor(a)
    u = Tensor.onehot(t, vocs)
    assert u.shape == t.shape + [vocs]
    b = np.array(u)
    exp = np.array([
        [[1., 0., 0., 0., 0.],   # 0
         [0., 0., 1., 0., 0.],   # 2
         [0., 0., 0., 0., 1.]],  # 4
        [[0., 0., 1., 0., 0.],   # 2
         [0., 0., 0., 1., 0.],   # 3
         [1., 0., 0., 0., 0.]]   # 0
    ], dtype=np.float32)
    assert np.allclose(b, exp)
